from rich.progress import Progress

from depscan.lib import config
from depscan.lib.logger import console, LOG
from depscan.lib.package_query.npm_pkg import npm_pkg_risk
from depscan.lib.package_query.pkg_query import get_lookup_url, httpclient
from depscan.lib.package_query.pypi_pkg import pypi_pkg_risk
from depscan.lib.package_query.cargo_pkg import cargo_pkg_risk


def metadata_from_registry(registry_type, scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query registry for the package metadata

    :param registry_type: The type of registry to query
    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return:  A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    metadata_dict = {}
    # Circuit breaker flag to break the risk audit in case of many api errors
    circuit_breaker = False
    # Track the api failures count
    failure_count = 0
    done_count = 0
    with Progress(
        console=console,
        transient=True,
        redirect_stderr=False,
        redirect_stdout=False,
        refresh_per_second=1,
        disable=len(pkg_list) < 10,
    ) as progress:
        task = progress.add_task("[green] Auditing packages", total=len(pkg_list))
        for pkg in pkg_list:
            if circuit_breaker:
                LOG.info(
                    "Risk audited has been interrupted due to frequent api "
                    "errors. Please try again later."
                )
                progress.stop()
                return {}
            scope = pkg.get("scope", "").lower()
            key, lookup_url = get_lookup_url(registry_type, pkg)
            if not key or not lookup_url or key.startswith("https://"):
                progress.advance(task)
                continue
            progress.update(task, description=f"Checking {key}")
            try:
                r = httpclient.get(
                    url=lookup_url,
                    follow_redirects=True,
                    timeout=config.request_timeout_sec,
                )
                json_data = r.json()
                # Npm is returning textual errors at times
                if isinstance(json_data, str):
                    continue
                # Npm returns this error if the package is not found
                if isinstance(json_data, dict) and (
                    json_data.get("code") == "MethodNotAllowedError"
                    or r.status_code > 400
                ):
                    continue
                is_private_pkg = False
                if private_ns:
                    namespace_prefixes = private_ns.split(",")
                    for ns in namespace_prefixes:
                        if key.lower().startswith(ns.lower()) or key.lower().startswith(
                            "@" + ns.lower()
                        ):
                            is_private_pkg = True
                            break
                risk_metrics = {}
                match registry_type:
                    case "npm":
                        risk_metrics = npm_pkg_risk(
                            json_data, is_private_pkg, scope, pkg
                        )
                    case "pypi":
                        project_type_pkg = f"python:{key}".lower()
                        required_pkgs = scoped_pkgs.get("required", [])
                        optional_pkgs = scoped_pkgs.get("optional", [])
                        excluded_pkgs = scoped_pkgs.get("excluded", [])
                        if (
                            pkg.get("purl") in required_pkgs
                            or project_type_pkg in required_pkgs
                        ):
                            scope = "required"
                        elif (
                            pkg.get("purl") in optional_pkgs
                            or project_type_pkg in optional_pkgs
                        ):
                            scope = "optional"
                        elif (
                            pkg.get("purl") in excluded_pkgs
                            or project_type_pkg in excluded_pkgs
                        ):
                            scope = "excluded"
                        risk_metrics = pypi_pkg_risk(
                            json_data, is_private_pkg, scope, pkg
                        )
                    case "cargo":
                        risk_metrics = cargo_pkg_risk(
                            json_data, is_private_pkg, scope, pkg
                        )
                    case _:
                        pass
                # TODO: remove unnecessary if elif statements
                # if registry_type == "npm":
                #     pass
                # elif registry_type == "pypi":
                #     pass
                metadata_dict[key] = {
                    "scope": scope,
                    "purl": pkg.get("purl"),
                    "pkg_metadata": json_data,
                    "risk_metrics": risk_metrics,
                    "is_private_pkg": is_private_pkg,
                }
            except Exception as e:
                LOG.debug(e)
                failure_count += 1
            progress.advance(task)
            done_count += 1
            if failure_count >= config.max_request_failures:
                circuit_breaker = True
    LOG.debug(
        "Retrieved package metadata for %d/%d packages. Failures count %d",
        done_count,
        len(pkg_list),
        failure_count,
    )
    return metadata_dict


def cargo_metadata(scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query cargo for the package metadata
    """
    return metadata_from_registry("cargo", scoped_pkgs, pkg_list, private_ns)


def npm_metadata(scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query npm for the package metadata

    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return: A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    return metadata_from_registry("npm", scoped_pkgs, pkg_list, private_ns)


def pypi_metadata(scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query pypi for the package metadata

    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return: A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    return metadata_from_registry("pypi", scoped_pkgs, pkg_list, private_ns)



"""
Module for parsing command line arguments and TOML configuration files.

This module provides a class, `ArgumentParser`, which extends the functionality
of `argparse.ArgumentParser` by allowing users to specify default values for
arguments in a TOML file, in addition to the command line.
"""
# Based on https://github.com/florianmahner/tomlparse/blob/main/tomlparse/argparse.py
# MIT license
import argparse
import os
from typing import Any, Dict, List, MutableMapping, Optional, Tuple

try:
    import tomllib
except ImportError:
    import tomli as tomllib


class ArgumentParser(argparse.ArgumentParser):
    """A wrapper of the argparse.ArgumentParser class that adds the ability to
    specify the values for arguments using a TOML file.

    This class extends the functionality of the standard argparse.ArgumentParser by allowing
    users to specify default values for arguments in a TOML file, in addition to the command line.
    We can use all functionalities from the argument parser as usual:

    Example:
        >>> from depscan.lib.tomlparse import argparse
        >>> parser = argparse.ArgumentParser(description='Example argparse-toml app')
        >>> parser.add_argument('--foo', type=int, help='An example argument')
        >>> args = parser.parse_args()

    The above code will work as with the standard argparse.ArgumentParser class. We can also
    specify the default values for the arguments in a TOML file. For this the TOML ArgumentParser
    has one additional argument: `--config`. The `--config` argument is used
    to specify the path to the TOML file.

    We have the following hierarchy of arguments:
        1. Arguments passed through the command line are selected over TOML
           arguments, even if both are passed
        2. Arguments from the TOML file are preferred over the default arguments
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        default_config = os.path.join(os.getcwd(), ".config", "depscan.toml")
        self.add_argument("--config", help="Path to the configuration file. Default: $PWD/.config/depscan.toml",
                          default=os.getenv("DEPSCAN_CONFIG", default_config))

    def extract_args(
        self, args: Optional[List[str]] = None, namespace: Optional[object] = None
    ) -> Tuple[argparse.Namespace, argparse.Namespace]:
        """Find the default arguments of the argument parser if any and the
        ones that are passed through the command line"""
        default_args = super().parse_args([])
        cmdl_args = super().parse_args(args, namespace)

        return default_args, cmdl_args

    def find_changed_args(
        self, default_args: argparse.Namespace, sys_args: argparse.Namespace
    ) -> List[str]:
        """Find the arguments that have been changed from the command
        line to replace the .toml arguments"""
        default_dict = vars(default_args)
        sys_dict = vars(sys_args)
        changed_dict = []
        for key, value in default_dict.items():
            sys_value = sys_dict[key]
            if sys_value != value:
                changed_dict.append(key)
        return changed_dict

    def load_toml(self, path: str) -> MutableMapping[str, Any]:
        try:
            with open(path, "rb") as f:
                config = tomllib.load(f)
        except FileNotFoundError:
            self.error(f'Configuration file "{path}" doesn\'t exist')
        return config

    def remove_nested_keys(self, dictionary: Dict[str, Any]) -> Dict[str, Any]:
        new_dict = {}
        for key, value in dictionary.items():
            if not isinstance(value, dict):
                new_dict[key] = value
        return new_dict

    def parse_args(
        self, args: Optional[List[str]] = None, namespace: Optional[object] = None
    ) -> argparse.Namespace:
        """Parse the arguments from the command line and the TOML file
        and return the updated arguments. Same functionality as the
        `argparse.ArgumentParser.parse_args` method."""
        default_args, sys_args = self.extract_args(args, namespace)
        config = sys_args.config
        # These are the default arguments options updated by the command line
        if not config or not os.path.exists(config):
            return sys_args

        # If a config file is passed, update the cmdl args with the config file unless
        # the argument is already specified in the command line
        toml_data = self.load_toml(config)
        changed_args = self.find_changed_args(default_args, sys_args)
        toml_args = self.remove_nested_keys(toml_data)

        # Replaced unchanged command line arguments with arguments from
        # the TOML file.
        for key, value in toml_args.items():
            if key not in changed_args:
                setattr(sys_args, key, value)
                # Support both hyphen and underscore representations
                setattr(sys_args, key.replace("-", "_"), value)

        return sys_args



import os
from depscan import get_version
from depscan.lib import tomlparse


def build_parser():
    parser = tomlparse.ArgumentParser(
        description="Fully open-source security and license audit for "
        "application dependencies and container images based on "
        "known vulnerabilities and advisories.",
        epilog="Visit https://github.com/owasp-dep-scan/dep-scan to learn more",
    )
    parser.add_argument(
        "--no-banner",
        action="store_true",
        default=False,
        dest="no_banner",
        help="Do not display the logo and donation banner. Please make a donation to OWASP before using this argument.",
    )
    parser.add_argument(
        "-i",
        "--src",
        default=os.getenv("DEPSCAN_SOURCE_DIR_IMAGE", os.getcwd()),
        dest="src_dir_image",
        help="Source directory or container image or binary file",
    )
    parser.add_argument(
        "-o",
        "--reports-dir",
        default=os.getenv("DEPSCAN_REPORTS_DIR", os.path.join(os.getcwd(), "reports")),
        dest="reports_dir",
        help="Reports directory",
    )
    parser.add_argument(
        "--csaf",
        action="store_true",
        default=False,
        dest="csaf",
        help="Generate a OASIS CSAF VEX document",
    )
    parser.add_argument(
        "--profile",
        default="generic",
        choices=(
            "appsec",
            "research",
            "operational",
            "threat-modeling",
            "license-compliance",
            "generic",
            "machine-learning",
            "ml",
            "deep-learning",
            "ml-deep",
            "ml-tiny",
        ),
        dest="profile",
        help="Profile to use while generating the BOM. For granular control, use the arguments --bom-engine, --vulnerability-analyzer, or --reachability-analyzer.",
    )
    parser.add_argument(
        "--lifecycle",
        choices=("pre-build", "build", "post-build"),
        nargs="+",
        type=str,
        dest="lifecycles",
        help="Product lifecycle for the generated BOM. Multiple values allowed.",
    )
    parser.add_argument(
        "--technique",
        choices=(
            "auto",
            "source-code-analysis",
            "binary-analysis",
            "manifest-analysis",
            "hash-comparison",
            "instrumentation",
            "filename",
        ),
        nargs="+",
        type=str,
        dest="techniques",
        help="Analysis technique to use for BOM generation. Multiple values allowed.",
    )
    engine_group = parser.add_mutually_exclusive_group(required=False)
    engine_group.add_argument(
        "--bom-engine",
        choices=(
            "auto",
            "CdxgenGenerator",
            "CdxgenServerGenerator",
            "CdxgenImageBasedGenerator",
            "BlintGenerator",
        ),
        default="auto",
        dest="bom_engine",
        help="BOM generation engine to use. Defaults to automatic selection based on project type and lifecycle.",
    )
    engine_group.add_argument(
        "--vulnerability-analyzer",
        choices=(
            "auto",
            "VDRAnalyzer",
            "LifecycleAnalyzer",
        ),
        default="auto",
        dest="vuln_analyzer",
        help="Vulnerability analyzer to use. Defaults to automatic selection based on bom_dir argument.",
    )
    parser.add_argument(
        "--reachability-analyzer",
        choices=(
            "off",
            "FrameworkReachability",
            "SemanticReachability",
        ),
        default="FrameworkReachability",
        dest="reachability_analyzer",
        help="Reachability analyzer to use. Default FrameworkReachability.",
    )
    parser.add_argument(
        "--no-suggest",
        action="store_false",
        default=True,
        dest="suggest",
        help="Disable suggest mode",
    )
    parser.add_argument(
        "--risk-audit",
        action="store_true",
        default=os.getenv("ENABLE_OSS_RISK", "") in ("true", "1"),
        dest="risk_audit",
        help="Perform package risk audit (slow operation). Npm only.",
    )
    parser.add_argument(
        "--cdxgen-args",
        default=os.getenv("CDXGEN_ARGS"),
        dest="cdxgen_args",
        help="Additional arguments to pass to cdxgen",
    )
    parser.add_argument(
        "--private-ns",
        dest="private_ns",
        default=os.getenv("PKG_PRIVATE_NAMESPACE"),
        help="Private namespace to use while performing oss risk audit. "
        "Private packages should not be available in public registries "
        "by default. Comma separated values accepted.",
    )
    parser.add_argument(
        "-t",
        "--type",
        nargs="+",
        type=str,
        dest="project_type",
        default=os.getenv("DEPSCAN_PROJECT_TYPE", "universal").split(","),
        help="Override project types if auto-detection is incorrect. Multiple values supported.",
    )
    bom_group = parser.add_mutually_exclusive_group(required=False)
    bom_group.add_argument(
        "--bom",
        dest="bom",
        help="Examine using the given Software Bill-of-Materials (SBOM) file "
        "in CycloneDX format. Use cdxgen command to produce one.",
    )
    bom_group.add_argument(
        "--bom-dir",
        dest="bom_dir",
        help="Examine all the Bill-of-Materials (BOM) files in the given directory.",
    )
    bom_group.add_argument(
        "--purl",
        dest="search_purl",
        help="Scan a single package url.",
    )
    parser.add_argument(
        "--report-template",
        dest="report_template",
        help="Jinja template file used for rendering a custom report",
    )
    parser.add_argument(
        "--report-name",
        default="rendered.report",
        dest="report_name",
        help="Filename of the custom report written to the --reports-dir",
    )
    parser.add_argument(
        "--deep",
        action="store_true",
        default=False,
        dest="deep_scan",
        help="Perform deep scan by passing this --deep argument to cdxgen. "
        "Useful while scanning docker images and OS packages.",
    )
    parser.add_argument(
        "--fuzzy-search",
        action="store_true",
        default=False,
        dest="fuzzy_search",
        help="Perform fuzzy search by creating variations of package names. Use this when the input SBOM lacks a PURL.",
    )
    parser.add_argument(
        "--search-order",
        choices=(
            "purl",
            "pcu",
            "cpe",
            "cpu",
            "url",
        ),
        default="pcu",
        dest="search_order",
        help="Attributes to use while searching for vulnerabilities. Default: PURL, CPE, URL (pcu).",
    )
    parser.add_argument(
        "--no-universal",
        action="store_true",
        default=False,
        dest="non_universal_scan",
        help="Depscan would attempt to perform a single universal scan "
        "instead of individual scans per language type.",
    )
    parser.add_argument(
        "--no-vuln-table",
        action="store_true",
        default=False,
        dest="no_vuln_table",
        help="Do not print the table with the full list of vulnerabilities. "
        "This can help reduce console output.",
    )
    parser.add_argument(
        "--server",
        action="store_true",
        default=False,
        dest="server_mode",
        help="Run depscan as a server",
    )
    parser.add_argument(
        "--server-host",
        default=os.getenv("DEPSCAN_HOST", "127.0.0.1"),
        dest="server_host",
        help="depscan server host",
    )
    parser.add_argument(
        "--server-port",
        default=os.getenv("DEPSCAN_PORT", "7070"),
        dest="server_port",
        help="depscan server port",
    )
    parser.add_argument(
        "--server-allowed-hosts",
        nargs="*",
        help="List of allowed hostnames or IPs that can access the server (e.g., 'localhost 192.168.1.10'). If unspecified, no host allowlist is enforced.",
        default=None,
    )

    parser.add_argument(
        "--server-allowed-paths",
        nargs="*",
        help="List of allowed filesystem paths that can be scanned by the server. Restricts `path` parameter in /scan requests.",
        default=None,
    )
    parser.add_argument(
        "--cdxgen-server",
        default=os.getenv("CDXGEN_SERVER_URL"),
        dest="cdxgen_server",
        help="cdxgen server url. Eg: http://cdxgen:9090",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        dest="enable_debug",
        help="Run depscan in debug mode.",
    )
    output_group = parser.add_mutually_exclusive_group(required=False)
    output_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        dest="quiet",
        help="Makes depscan quiet.",
    )
    output_group.add_argument(
        "--explain",
        action="store_true",
        default=False,
        dest="explain",
        help="Makes depscan to explain the various analysis. Useful for creating detailed reports.",
    )
    parser.add_argument(
        "--explanation-mode",
        choices=(
            "Endpoints",
            "EndpointsAndReachables",
            "NonReachables",
            "LLMPrompts",
        ),
        default="EndpointsAndReachables",
        dest="explanation_mode",
        help="Style of explanation needed. Defaults to Endpoints and Reachables.",
    )
    parser.add_argument(
        "--annotate",
        action="store_true",
        default=False,
        dest="annotate",
        help="Include the generated text VDR report as an annotation. Defaults to true when explain is enabled; false otherwise.",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Display the version",
        action="version",
        version="%(prog)s " + get_version(),
    )
    return parser



import ast
import os
import re

from custom_json_diff.lib.utils import file_read, file_write, json_load
from jinja2 import Environment

from depscan.lib.config import ignore_directories
from depscan.lib.logger import LOG

LIC_SYMBOL_REGEX = re.compile(r"[(),]")


def filter_ignored_dirs(dirs):
    """
    Method to filter directory list to remove ignored directories

    :param dirs: Directories to ignore
    :return: Filtered directory list
    """
    [
        dirs.remove(d)
        for d in list(dirs)
        if d.lower() in ignore_directories or d.startswith(".")
    ]
    return dirs


def find_python_reqfiles(path):
    """
    Method to find python requirements files

    :param path: Project directory
    :return: List of python requirement files
    """
    result = []
    req_files = [
        "requirements.txt",
        "Pipfile",
        "poetry.lock",
        "Pipfile.lock",
        "conda.yml",
        "pyproject.toml",
    ]
    for root, dirs, files in os.walk(path):
        filter_ignored_dirs(dirs)
        result.extend(os.path.join(root, name) for name in req_files if name in files)
    return result


def find_files(src, src_ext_name, quick=False, filter_dirs=True):
    """
    Method to find files with given extension

    :param src: source directory to search
    :param src_ext_name: type of source file
    :param quick: only return first match found
    :param filter_dirs: filter out ignored directories
    """
    result = []
    for root, dirs, files in os.walk(src):
        if filter_dirs:
            filter_ignored_dirs(dirs)
        for file in files:
            if file == src_ext_name or file.endswith(src_ext_name):
                result.append(os.path.join(root, file))
                if quick:
                    return result
    return result


def is_binary_string(content):
    """
    Method to check if the given content is a binary string
    """
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    return bool(content.translate(None, textchars))


def is_exe(src):
    """
    Detect if the source is a binary file

    :param src: Source path
    :return: True if binary file. False otherwise.
    """
    if os.path.isfile(src):
        try:
            with open(src, "rb") as f:
                return is_binary_string(f.read(1024))
        except Exception:
            return False
    return False


def detect_project_type(src_dir):
    """Detect project type by looking for certain files

    :param src_dir: Source directory
    :return List of detected types
    """
    # container image support
    if (
        "docker.io" in src_dir
        or "quay.io" in src_dir
        or ":latest" in src_dir
        or "@sha256" in src_dir
        or src_dir.endswith(".tar")
        or src_dir.endswith(".tar.gz")
    ):
        return ["docker"]
    # Check if the source is an exe file. Assume go for all binaries for now
    if is_exe(src_dir):
        return ["go", "binary"]
    project_types = []
    if find_python_reqfiles(src_dir) or find_files(src_dir, ".py", quick=True):
        project_types.append("python")
    if find_files(src_dir, "pom.xml", quick=True) or find_files(
        src_dir, ".gradle", quick=True
    ):
        project_types.append("java")
    if find_files(src_dir, ".gradle.kts", quick=True):
        project_types.append("kotlin")
    if find_files(src_dir, "build.sbt", quick=True):
        project_types.append("scala")
    if (
        find_files(src_dir, "package.json", quick=True)
        or find_files(src_dir, "yarn.lock", quick=True)
        or find_files(src_dir, "rush.json", quick=True)
    ):
        project_types.append("nodejs")
    if find_files(src_dir, "go.sum", quick=True) or find_files(
        src_dir, "Gopkg.lock", quick=True
    ):
        project_types.append("go")
    if find_files(src_dir, "Cargo.lock", quick=True):
        project_types.append("rust")
    if find_files(src_dir, "composer.json", quick=True):
        project_types.append("php")
    if find_files(src_dir, ".csproj", quick=True):
        project_types.append("dotnet")
    if find_files(src_dir, "Gemfile", quick=True) or find_files(
        src_dir, "Gemfile.lock", quick=True
    ):
        project_types.append("ruby")
    if find_files(src_dir, "deps.edn", quick=True) or find_files(
        src_dir, "project.clj", quick=True
    ):
        project_types.append("clojure")
    if find_files(src_dir, "conan.lock", quick=True) or find_files(
        src_dir, "conanfile.txt", quick=True
    ):
        project_types.append("cpp")
    if find_files(src_dir, "pubspec.lock", quick=True) or find_files(
        src_dir, "pubspec.yaml", quick=True
    ):
        project_types.append("dart")
    if find_files(src_dir, "cabal.project.freeze", quick=True):
        project_types.append("haskell")
    if find_files(src_dir, "mix.lock", quick=True):
        project_types.append("elixir")
    if find_files(
        os.path.join(src_dir, ".github", "workflows"),
        ".yml",
        quick=True,
        filter_dirs=False,
    ):
        project_types.append("github")
    # jars
    if "java" not in project_types and find_files(src_dir, ".jar", quick=True):
        project_types.append("jar")
    # Jenkins plugins or plain old jars
    if "java" not in project_types and find_files(src_dir, ".hpi", quick=True):
        project_types.append("jenkins")
    if find_files(src_dir, ".yml", quick=True) or find_files(
        src_dir, ".yaml", quick=True
    ):
        project_types.append("yaml-manifest")
    return project_types


def cleanup_license_string(license_str):
    """
    Method to clean up license string by removing problematic symbols and
    making certain keywords consistent

    :param license_str: String to clean up
    :return: Cleaned up version
    """
    if not license_str:
        license_str = ""
    license_str = (
        license_str.replace(" / ", " OR ")
        .replace("/", " OR ")
        .replace(" & ", " OR ")
        .replace("&", " OR ")
    )
    license_str = LIC_SYMBOL_REGEX.sub("", license_str)
    return license_str.upper()


def get_all_imports(src_dir):
    """
    Method to collect all package imports from a python file
    No longer required since cdxgen does python analysis already
    """
    import_list = set()
    py_files = find_files(src_dir, ".py")
    if not py_files:
        return import_list
    for afile in py_files:
        parsed = ast.parse(file_read(os.path.join(afile), True, log=LOG))
        for node in ast.walk(parsed):
            if isinstance(node, ast.Import):
                for name in node.names:
                    pkg = name.name.split(".")[0]
                    import_list.add(pkg)
                    import_list.add(pkg.lower().replace("py", ""))
            elif isinstance(node, ast.ImportFrom):
                if node.level > 0:
                    continue
                if getattr(node, "module"):
                    if node.module:
                        pkg = node.module.split(".")[0]
                        import_list.add(pkg)
                        import_list.add(pkg.lower().replace("py", ""))
    return import_list


def render_template_report(
    vdr_file,
    bom_file,
    pkg_vulnerabilities,
    pkg_group_rows,
    summary,
    template_file,
    result_file,
    depscan_options={},
):
    """
    Render the given vdr_file (falling back to bom_file if no vdr was written)
    and summary dict using the template_file with Jinja, rendered output is written
    to named result_file in reports directory.
    """
    bom = {}
    if vdr_file:
        bom = json_load(vdr_file, log=LOG)
    if not bom:
        bom = json_load(bom_file, log=LOG)
    template = file_read(template_file, log=LOG)
    jinja_env = Environment(autoescape=True)
    jinja_tmpl = jinja_env.from_string(template)
    report_result = jinja_tmpl.render(
        metadata=bom.get("metadata"),
        vulnerabilities=bom.get("vulnerabilities"),
        components=bom.get("components"),
        dependencies=bom.get("dependencies"),
        services=bom.get("services"),
        summary=summary,
        pkg_vulnerabilities=pkg_vulnerabilities,
        pkg_group_rows=pkg_group_rows,
    )
    file_write(
        result_file,
        report_result,
        error_msg=f"Failed to export report: {result_file}",
        success_msg=f"Report written to {result_file}.",
        log=LOG,
    )



import math
import os

from depscan.lib import config
from depscan.lib.logger import LOG

try:
    from hishel.httpx import SyncCacheClient

    httpclient = SyncCacheClient()
except ImportError:
    import httpx

    httpclient = httpx


def get_lookup_url(registry_type, pkg):
    """
    Generating the lookup URL based on the registry type and package
    information.

    :param registry_type: The type of registry ("npm" or "pypi")
    :param pkg: Dict or string of package information
    :returns: Package name, lookup URL
    """
    vendor = None
    if isinstance(pkg, dict):
        vendor = pkg.get("vendor")
        name = pkg.get("name")
    else:
        tmp_a = pkg.split("|")
        name = tmp_a[len(tmp_a) - 2]
        if len(tmp_a) == 3:
            vendor = tmp_a[0]
    key = name
    # Prefix vendor for npm
    if registry_type == "npm":
        if vendor and vendor != "npm":
            # npm expects namespaces to start with an @
            if not vendor.startswith("@"):
                vendor = "@" + vendor
            key = f"{vendor}/{name}"
        return key, f"{config.NPM_SERVER}/{key}"
    if registry_type == "pypi":
        return key, f"{config.PYPI_SERVER}/{key}/json"
    if registry_type == "cargo":
        return key, f"{config.CARGO_SERVER}/{key}"
    return None, None


def get_category_score(
    param, max_value=config.DEFAULT_MAX_VALUE, weight=config.DEFAULT_WEIGHT
):
    """
    Return parameter score given its current value, max value and
    parameter weight.

    :param param: The current value of the parameter
    :param max_value: The maximum value of the parameter
    :param weight: The weight of the parameter
    :return: The calculated score as a float value
    """
    try:
        param = float(param)
    except ValueError:
        param = 0
    try:
        max_value = float(max_value)
    except ValueError:
        max_value = config.DEFAULT_MAX_VALUE
    try:
        weight = float(weight)
    except ValueError:
        weight = config.DEFAULT_WEIGHT
    return (
        0
        if weight == 0 or math.log(1 + max(param, max_value)) == 0
        else (math.log(1 + param) / math.log(1 + max(param, max_value))) * weight
    )


def calculate_risk_score(risk_metrics):
    """
    Method to calculate a total risk score based on risk metrics. This is
    based on a weighted formula and might require customization based on use
    cases

    :param risk_metrics: Dict containing many risk metrics
    :return: The calculated total risk score
    """
    if not risk_metrics:
        return 0
    num_risks = 0
    working_score = 0
    total_weight = 0
    for k, v in risk_metrics.items():
        # Is the _risk key set to True
        if k.endswith("_risk") and v is True:
            risk_category = k.replace("_risk", "")
            risk_category_value = risk_metrics.get(f"{risk_category}_value", 0)
            risk_category_max = getattr(
                config, f"{risk_category}_max", config.DEFAULT_MAX_VALUE
            )
            risk_category_weight = getattr(
                config, f"{risk_category}_weight", config.DEFAULT_WEIGHT
            )
            risk_category_base = getattr(config, f"{risk_category}", 0)
            value = risk_category_value
            if (
                risk_category_base
                and (
                    isinstance(risk_category_base, float)
                    or isinstance(risk_category_base, int)
                )
                and risk_category_base > risk_category_value
            ):
                value = risk_category_base - risk_category_value
            elif risk_category_max and risk_category_max > risk_category_value:
                value = risk_category_max - risk_category_value
            cat_score = get_category_score(
                value, risk_category_max, risk_category_weight
            )
            total_weight += risk_category_weight
            working_score += min(cat_score, 1)
            num_risks += 1
    working_score = round(working_score * total_weight / config.total_weight, 5)
    working_score = max(min(working_score, 1), 0)
    return working_score


def compute_time_risks(
    risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
):
    """
    Compute risks based on creation, modified and time elapsed

    :param risk_metrics: A dict containing the risk metrics for the package.
    :param created_now_diff: Time difference from creation of the package and
    the current time.
    :param mod_create_diff: Time difference from
    modification and creation of the package.
    :param latest_now_diff: Time difference between the latest version of the
    package and the current
    time.
    :return: The updated risk metrics dictionary with the calculated
    risks and values.
    """
    # Check if the package is at least 1 year old. Quarantine period.
    if created_now_diff.total_seconds() < config.created_now_quarantine_seconds:
        risk_metrics["created_now_quarantine_seconds_risk"] = True
        risk_metrics["created_now_quarantine_seconds_value"] = (
            latest_now_diff.total_seconds()
        )

    # Check for the maximum seconds difference between latest version and now
    if latest_now_diff.total_seconds() > config.latest_now_max_seconds:
        risk_metrics["latest_now_max_seconds_risk"] = True
        risk_metrics["latest_now_max_seconds_value"] = latest_now_diff.total_seconds()
        # Since the package is quite old we can relax the min versions risk
        risk_metrics["pkg_min_versions_risk"] = False
    else:
        # Check for the minimum seconds difference between creation and
        # modified date This check catches several old npm packages that was
        # created and immediately updated within a day To reduce noise we
        # check for the age first and perform this check only for newish
        # packages
        if mod_create_diff.total_seconds() < config.mod_create_min_seconds:
            risk_metrics["mod_create_min_seconds_risk"] = True
            risk_metrics["mod_create_min_seconds_value"] = (
                mod_create_diff.total_seconds()
            )
    # Check for the minimum seconds difference between latest version and now
    if latest_now_diff.total_seconds() < config.latest_now_min_seconds:
        risk_metrics["latest_now_min_seconds_risk"] = True
        risk_metrics["latest_now_min_seconds_value"] = latest_now_diff.total_seconds()
    return risk_metrics



#!/usr/bin/env python3 -W ignore::DeprecationWarning
# -*- coding: utf-8 -*-

import contextlib
import os
import sys
from typing import List

from analysis_lib import (
    ReachabilityAnalysisKV,
    VdrAnalysisKV,
)
from analysis_lib.csaf import export_csaf, write_toml
from analysis_lib.search import get_pkgs_by_scope
from analysis_lib.utils import (
    get_all_bom_files,
    get_all_pkg_list,
    get_pkg_list,
    licenses_risk_table,
    pkg_risks_table,
    summary_stats,
)
from analysis_lib.vdr import VDRAnalyzer
from analysis_lib.reachability import get_reachability_impl
from custom_json_diff.lib.utils import json_load
from rich.panel import Panel
from rich.terminal_theme import DEFAULT_TERMINAL_THEME, MONOKAI
from vdb.lib import config
from vdb.lib import db6 as db_lib
from vdb.lib.utils import parse_purl

from depscan import get_version
from depscan.cli_options import build_parser
from depscan.lib import explainer, utils
from depscan.lib.audit import audit, risk_audit, risk_audit_map, type_audit_map
from depscan.lib.bom import (
    annotate_vdr,
    create_empty_vdr,
    create_bom,
    export_bom,
    get_pkg_by_type,
)
from depscan.lib.config import (
    DEPSCAN_DEFAULT_VDR_FILE,
    UNIVERSAL_SCAN_TYPE,
    VDB_AGE_HOURS,
    license_data_dir,
    pkg_max_risk_score,
    spdx_license_list,
    vdb_database_url,
)
from depscan.lib.license import build_license_data, bulk_lookup
from depscan.lib.logger import DEBUG, LOG, SPINNER, console, IS_CI

from reporting_lib.htmlgen import ReportGenerator

if sys.platform == "win32" and os.environ.get("PYTHONIOENCODING") is None:
    sys.stdin.reconfigure(encoding="utf-8")
    sys.stdout.reconfigure(encoding="utf-8")
    sys.stderr.reconfigure(encoding="utf-8")

LOGO = """
  _|  _  ._   _  _  _. ._
 (_| (/_ |_) _> (_ (_| | |
         |
"""

SERVER_LIB = None
try:
    from server_lib import simple, ServerOptions

    SERVER_LIB = simple
except ImportError:
    pass

ORAS_AVAILABLE = False
try:
    from vdb.lib.orasclient import download_image

    ORAS_AVAILABLE = True
except ImportError:
    pass


def build_args():
    """
    Constructs command line arguments for the depscan tool
    """
    parser = build_parser()
    return parser.parse_args()


def vdr_analyze_summarize(
    project_type,
    results,
    suggest_mode,
    scoped_pkgs,
    bom_file,
    bom_dir,
    reports_dir,
    pkg_list,
    reachability_analyzer,
    reachability_options,
    no_vuln_table=False,
    fuzzy_search=False,
    search_order=None,
):
    """
    Method to perform VDR analysis followed by summarization.
    :param project_type: Project type.
    :param results: Scan or audit results.
    :param suggest_mode: Normalize fix versions automatically.
    :param scoped_pkgs: Dict containing package scopes.
    :param bom_file: Single BOM file.
    :param bom_dir: Directory containining bom files.
    :param reports_dir: Directory containining report files.
    :param pkg_list: Direct list of packages when the bom file is empty.
    :param reachability_analyzer: Reachability Analyzer specified.
    :param reachability_options: Reachability Analyzer options.
    :param no_vuln_table: Boolean to indicate if the results should get printed
            to the console.
    :param fuzzy_search: Perform fuzzy search.
    :param search_order: Search order.

    :return: A dict of vulnerability and severity summary statistics
    """
    pkg_vulnerabilities = []
    summary = {}
    direct_purls = {}
    reached_purls = {}
    reached_services = {}
    endpoint_reached_purls = {}
    # Perform the reachability analysis first
    reach_result = get_reachability_impl(
        reachability_analyzer, reachability_options
    ).process()
    # We now have reachability results, OpenAPI endpoints, BOMs, and component scope information.
    if reach_result and reach_result.success:
        direct_purls = reach_result.direct_purls
        reached_purls = reach_result.reached_purls
        reached_services = reach_result.reached_services
        endpoint_reached_purls = reach_result.endpoint_reached_purls
    console.record = True
    # We might already have the needed slices files when we reach here.
    options = VdrAnalysisKV(
        project_type,
        results,
        pkg_aliases={},
        purl_aliases={},
        suggest_mode=suggest_mode,
        scoped_pkgs=scoped_pkgs,
        no_vuln_table=no_vuln_table,
        bom_file=bom_file,
        bom_dir=bom_dir,
        pkg_list=pkg_list,
        direct_purls=direct_purls,
        reached_purls=reached_purls,
        reached_services=reached_services,
        endpoint_reached_purls=endpoint_reached_purls,
        console=console,
        logger=LOG,
        fuzzy_search=fuzzy_search,
        search_order=search_order,
    )
    ds_version = get_version()
    vdr_result = VDRAnalyzer(vdr_options=options).process()
    # Set vdr_file in report folder
    vdr_file = (
        os.path.join(reports_dir, os.path.basename(bom_file)) if bom_file else None
    )
    vdr_file = vdr_file.replace(".cdx.json", ".vdr.json") if vdr_file else None
    if not vdr_file and bom_dir:
        vdr_file = os.path.join(bom_dir, DEPSCAN_DEFAULT_VDR_FILE)
    if vdr_result.success:
        pkg_vulnerabilities = vdr_result.pkg_vulnerabilities
        cdx_vdr_data = None
        # Always create VDR files even when empty
        if pkg_vulnerabilities is not None:
            # Case 1: Single BOM file resulting in a single VDR file
            if bom_file:
                cdx_vdr_data = json_load(bom_file, log=LOG)
            # Case 2: Multiple BOM files in a bom directory
            elif bom_dir:
                cdx_vdr_data = create_empty_vdr(pkg_list, ds_version)
        if cdx_vdr_data:
            export_bom(cdx_vdr_data, ds_version, pkg_vulnerabilities, vdr_file)
            LOG.debug(f"The VDR file '{vdr_file}' was created successfully.")
        else:
            LOG.debug(
                f"VDR file '{vdr_file}' was not created for the type {project_type}."
            )
        summary = summary_stats(pkg_vulnerabilities)
    elif bom_dir or bom_file or pkg_list:
        if project_type != "bom":
            LOG.info("No vulnerabilities found for project type '%s'!", project_type)
        else:
            LOG.info("No vulnerabilities found!")
    return summary, vdr_file, vdr_result


def set_project_types(args, src_dir):
    """
    Detects the project types and perform the right type of scan

    :param args: cli arguments
    :param src_dir: source directory

    :return: A tuple containing the package list, the parsed package URL object,
    and the list of project types.
    """
    pkg_list, purl_obj = [], {}
    project_types_list: List[str] = []
    if args.search_purl:
        purl_obj = parse_purl(args.search_purl)
        purl_obj["purl"] = args.search_purl
        purl_obj["vendor"] = purl_obj.get("namespace")
        if purl_obj.get("type"):
            project_types_list = [purl_obj.get("type", "")]
        pkg_list = [purl_obj]
    elif args.bom or args.bom_dir:
        project_types_list = ["bom"]
    elif args.project_type:
        project_types_list = (
            args.project_type
            if isinstance(args.project_type, list)
            else args.project_type.split(",")
        )
        if len(project_types_list) == 1 and "," in project_types_list[0]:
            project_types_list = project_types_list[0].split(",")
    elif not args.non_universal_scan:
        project_types_list = [UNIVERSAL_SCAN_TYPE]
    else:
        project_types_list = utils.detect_project_type(src_dir)
    return pkg_list, project_types_list


def run_depscan(args):
    """
    Detects the project type, performs various scans and audits,
    and generates reports based on the results.
    """
    perform_risk_audit = args.risk_audit
    # declare variables that get initialized only conditionally
    (
        summary,
        vdr_file,
        bom_file,
        prebuild_bom_file,
        build_bom_file,
        postbuild_bom_file,
        container_bom_file,
        operations_bom_file,
        pkg_list,
        all_pkg_vulnerabilities,
        all_pkg_group_rows,
    ) = (None, None, None, None, None, None, None, None, None, [], {})
    if (
        os.getenv("CI")
        and not os.getenv("GITHUB_REPOSITORY", "").lower().startswith("owasp")
        and not args.no_banner
        and not os.getenv("INPUT_THANK_YOU", "") == "I have sponsored OWASP-dep-scan."
    ):
        console.print(
            Panel(
                "OWASP foundation relies on donations to fund our projects.\nPlease donate at: https://owasp.org/donate/?reponame=www-project-dep-scan&title=OWASP+depscan",
                title="Donate to the OWASP Foundation",
                expand=False,
            )
        )
    # Should we be quiet
    if args.quiet:
        args.explain = False
        LOG.disabled = True
        args.enable_debug = False
        os.environ["SCAN_DEBUG_MODE"] = "off"
        os.environ["CDXGEN_DEBUG_MODE"] = "off"
        console.quiet = True
        args.no_vuln_table = True
    # Should we enable debug
    if args.enable_debug:
        os.environ["SCAN_DEBUG_MODE"] = "debug"
        os.environ["CDXGEN_DEBUG_MODE"] = "debug"
        LOG.setLevel(DEBUG)
    if args.server_mode:
        if SERVER_LIB:
            server_options = ServerOptions(
                server_host=args.server_host,
                server_port=args.server_port,
                cdxgen_server=args.cdxgen_server,
                allowed_hosts=args.server_allowed_hosts,
                allowed_paths=args.server_allowed_paths,
                console=console,
                logger=LOG,
                debug=args.enable_debug or os.environ.get("SCAN_DEBUG_MODE") == "debug",
                create_bom=create_bom,
                max_content_length=os.getenv("DEPSCAN_SERVER_MAX_CONTENT_LENGTH"),
            )
            return simple.run_server(server_options)
        else:
            LOG.info(
                "The required packages for server mode are unavailable. Reinstall depscan using `pip install owasp-depscan[all]`."
            )
            return False
    if not args.no_banner:
        with contextlib.suppress(UnicodeEncodeError):
            print(LOGO)
    # Break early if the user prefers CPE-based searches
    search_order = args.search_order
    if search_order:
        if search_order.startswith("c") and not args.bom and not args.bom_dir:
            LOG.warning(
                "To perform CPE-based searches, the SBOM must include a CPE identifier for each component. Generate the SBOM using a compatible tool such as Syft or Trivy, and invoke depscan with the --bom or --bom-dir argument."
            )
            LOG.info(
                "Alternatively, run depscan without the `--search-order` argument to perform PURL-based searches. This method is more accurate and recommended."
            )
            sys.exit(1)
        elif search_order.startswith("u") and not os.getenv("FETCH_LICENSE"):
            LOG.warning(
                "To perform URL-based searches, the SBOM must include externalReferences with a URL. Set the environment variable `FETCH_LICENSE=true` to force cdxgen to populate this attribute."
            )
            LOG.info(
                "Alternatively, include the project type `-t license` to ensure this attribute is populated."
            )
    src_dir = args.src_dir_image
    if not src_dir or src_dir == ".":
        if src_dir == "." or args.search_purl:
            src_dir = os.getcwd()
        # Try to infer from the bom file
        elif args.bom and os.path.exists(args.bom):
            src_dir = os.path.dirname(os.path.realpath(args.bom))
        elif args.bom_dir and os.path.exists(args.bom_dir):
            src_dir = os.path.realpath(args.bom_dir)
        else:
            src_dir = os.getcwd()
    reports_dir = args.reports_dir
    # User has not provided an explicit reports_dir. Reuse the bom_dir
    if not reports_dir and args.bom_dir:
        reports_dir = os.path.realpath(args.bom_dir)
    # Are we running for a BOM directory
    bom_dir_mode = args.bom_dir and os.path.exists(args.bom_dir)
    # Are we running with a config file
    config_file_mode = args.config and os.path.exists(args.config)
    depscan_options = {**vars(args), "src_dir": src_dir, "reports_dir": reports_dir}
    # Is the user looking for semantic analysis?
    # We can default to this when run against a BOM directory
    if (
        args.reachability_analyzer == "SemanticReachability"
    ) and args.vuln_analyzer != "LifecycleAnalyzer":
        LOG.debug(
            "Automatically switching to the `LifecycleAnalyzer` for vulnerability analysis."
        )
        depscan_options["vuln_analyzer"] = "LifecycleAnalyzer"
        args.vuln_analyzer = "LifecycleAnalyzer"
    # Should we download the latest vdb.
    if db_lib.needs_update(
        days=0,
        hours=VDB_AGE_HOURS,
        default_status=db_lib.get_db_file_metadata is not None,
    ):
        if ORAS_AVAILABLE:
            with console.status(
                f"Downloading the latest vulnerability database to {config.DATA_DIR}. Please wait ...",
                spinner=SPINNER,
            ) as vdb_download_status:
                if not IS_CI:
                    vdb_download_status.stop()
                # This line may exit with an exception if the database cannot be downloaded.
                # Example: urllib3.exceptions.IncompleteRead, urllib3.exceptions.ProtocolError, requests.exceptions.ChunkedEncodingError
                download_image(vdb_database_url, config.DATA_DIR)
        else:
            LOG.warning(
                "The latest vulnerability database is not found. Follow the documentation to manually download it."
            )
    if args.csaf:
        toml_file_path = os.getenv(
            "DEPSCAN_CSAF_TEMPLATE", os.path.join(src_dir, "csaf.toml")
        )
        if not os.path.exists(toml_file_path):
            LOG.info("CSAF toml not found, creating template in %s", src_dir)
            write_toml(toml_file_path)
            LOG.info("Please fill out the toml with your details and rerun depscan.")
            LOG.info(
                "Check out our CSAF documentation for an explanation of "
                "this feature. https://github.com/owasp-dep-scan/dep-scan"
                "/blob/master/contrib/CSAF_README.md"
            )
            LOG.info(
                "If you're just checking out how our generator works, "
                "feel free to skip filling out the toml and just rerun "
                "depscan."
            )
            sys.exit(0)
    pkg_list, project_types_list = set_project_types(args, src_dir)
    if args.search_purl:
        # Automatically enable risk audit for single purl searches
        perform_risk_audit = True
    # Construct the various report files
    html_report_file = depscan_options.get(
        "html_report_file", os.path.join(reports_dir, "depscan.html")
    )
    txt_report_file = depscan_options.get(
        "txt_report_file", os.path.join(reports_dir, "depscan.txt")
    )
    llm_prompt_file = depscan_options.get(
        "llm_prompt_file", os.path.join(reports_dir, "depscan-prompts.md")
    )
    run_config_file = os.path.join(reports_dir, "depscan.toml.sample")
    depscan_options["html_report_file"] = html_report_file
    depscan_options["txt_report_file"] = txt_report_file
    depscan_options["llm_prompt_file"] = llm_prompt_file
    # Create reports directory
    if reports_dir and not os.path.exists(reports_dir):
        os.makedirs(reports_dir, exist_ok=True)
    if len(project_types_list) > 1:
        LOG.debug("Multiple project types found: %s", project_types_list)
    # Enable license scanning
    if "license" in project_types_list or "license" in args.profile:
        os.environ["FETCH_LICENSE"] = "true"
        project_types_list.remove("license")
        console.print(
            Panel(
                "License audit is enabled for this scan. This would increase "
                "the time by up to 10 minutes.",
                title="License Audit",
                expand=False,
            )
        )
    # Let’s create a sample configuration file based on the CLI options used.
    if not config_file_mode:
        run_config = {**depscan_options}
        del run_config["no_banner"]
        write_toml(run_config_file, run_config, write_version=False)
        LOG.debug(
            f"Created a sample depscan config file at '{run_config_file}', based on this run."
        )
    # We have everything needed to start the composition analysis. There are many approaches to implementing an SCA tool.
    # Our style of analysis is comparable to that of an intelligent Hubble telescope or a rover—examining the same subject through multiple optics, colors, and depths to gain a deeper understanding.
    # We begin by iterating over the project types provided or assumed.
    for project_type in project_types_list:
        results = []
        vuln_analyzer = args.vuln_analyzer
        # Are we performing a lifecycle analysis
        if not args.search_purl and (
            vuln_analyzer == "LifecycleAnalyzer"
            or (vuln_analyzer == "auto" and bom_dir_mode)
        ):
            if args.reachability_analyzer == "SemanticReachability":
                if not args.bom_dir:
                    LOG.info(
                        "Semantic Reachability analysis requested for project type '%s'. This might take a while ...",
                        project_type,
                    )
                else:
                    LOG.info(
                        "Attempting semantic analysis using existing data at '%s'",
                        args.bom_dir,
                    )
            else:
                LOG.info(
                    "Lifecycle-based vulnerability analysis requested for project type '%s'. This might take a while ...",
                    project_type,
                )
            prebuild_bom_file = os.path.join(
                reports_dir, f"sbom-prebuild-{project_type}.cdx.json"
            )
            build_bom_file = os.path.join(
                reports_dir, f"sbom-build-{project_type}.cdx.json"
            )
            postbuild_bom_file = os.path.join(
                reports_dir, f"sbom-postbuild-{project_type}.cdx.json"
            )
            # We support only one container SBOM per project.
            # Projects that rely on docker compose and multiple services require some thinking
            container_bom_file = os.path.join(
                reports_dir, f"sbom-container-{project_type}.cdx.json"
            )
            operations_bom_file = os.path.join(
                reports_dir, f"sbom-operations-{project_type}.cdx.json"
            )
            if vuln_analyzer == "auto":
                vuln_analyzer = "LifecycleAnalyzer"
                depscan_options["vuln_analyzer"] = "LifecycleAnalyzer"
            # We need to set the following two values to make the rest of the code correctly use
            # the generated BOM files after lifecycle analysis
            depscan_options["lifecycle_analysis_mode"] = True
            if not args.bom_dir:
                args.bom_dir = os.path.realpath(reports_dir)
        # If the user opts out of lifecycle analysis, we need to maintain multiple SBOMs based on the project type.
        bom_file = os.path.join(reports_dir, f"sbom-{project_type}.cdx.json")
        risk_report_file = os.path.join(
            reports_dir, f"depscan-risk-{project_type}.json"
        )
        # Are we scanning a single purl
        if args.search_purl:
            bom_file = None
            creation_status = True
        # Are we scanning a bom file
        ###################
        # Note to students and researchers benchmarking depscan:
        #   we’ve seen attempts to run depscan using SBOMs generated by tools like Syft, Trivy, etc.
        # It’s important to understand that not all SBOMs contain the same level of detail.
        # Component PURLs can differ slightly, especially in qualifiers.
        #
        # For container SBOMs, qualifiers like distro_name and distro_version are critical for accurate results.
        # Tools like Syft and Trivy often include internal metadata—such as vendor IDs or fabricated CPE strings—to brute-force vulnerability matches.
        # Because of these inconsistencies, it’s not possible to achieve identical results with depscan when using a non-cdxgen or non-blint SBOM.
        # If in doubt, speak to us before benchmarking depscan. Don’t run depscan with default settings and expect magic.
        # SCA and xBOM are complex domains that require understanding, configuration, and continuous learning.
        ###################
        elif args.bom and os.path.exists(args.bom):
            bom_file = args.bom
            creation_status = True
        # Are we scanning a bom directory
        elif bom_dir_mode:
            bom_file = None
            creation_status = True
        else:
            # Create a bom for each project type
            creation_status = create_bom(
                bom_file,
                src_dir,
                {
                    **depscan_options,
                    "project_type": [project_type],
                    "bom_file": bom_file,
                    "prebuild_bom_file": prebuild_bom_file,
                    "build_bom_file": build_bom_file,
                    "postbuild_bom_file": postbuild_bom_file,
                    "container_bom_file": container_bom_file,
                    "operations_bom_file": operations_bom_file,
                },
            )
        if not creation_status:
            LOG.warning(
                "The BOM file `%s` was not created successfully. Set the `SCAN_DEBUG_MODE=debug` environment variable to troubleshoot.",
                bom_file,
            )
            continue
        # We have a BOM directory. Let’s aggregate all packages from every file within it.
        if args.bom_dir:
            LOG.debug(
                "Collecting components from all the BOM files at %s",
                args.bom_dir,
            )
            pkg_list = get_all_pkg_list(args.bom_dir)
        # We are working with a single BOM file and will collect all packages from it accordingly.
        elif bom_file:
            LOG.debug("Scanning using the bom file %s", bom_file)
            if not args.bom:
                LOG.info(
                    "To improve performance, cache the bom file and invoke "
                    "depscan with --bom %s instead of -i",
                    bom_file,
                )
            pkg_list, _ = get_pkg_list(bom_file)
        if not pkg_list and not args.bom_dir:
            LOG.info(
                "No packages were found in the project. Try generating the BOM manually or use the `CdxgenImageBasedGenerator` engine."
            )
            continue
        # Depending on the SBOM tool used, there may be details about component usage and scopes. Let’s analyze and interpret that information.
        scoped_pkgs = get_pkgs_by_scope(pkg_list)
        # Is the user interested in seeing license risks? Handle that first before any security-related analysis.
        if (
            os.getenv("FETCH_LICENSE", "") in (True, "1", "true")
            or "license" in args.profile
        ):
            licenses_results = bulk_lookup(
                build_license_data(license_data_dir, spdx_license_list),
                pkg_list=pkg_list,
            )
            license_report_file = os.path.join(
                reports_dir, f"license-{project_type}.json"
            )
            ltable = licenses_risk_table(
                project_type, licenses_results, license_report_file
            )
            if ltable and not args.no_vuln_table:
                console.print(ltable)
        # Do we support OSS risk audit for this type? If yes, proceed with the relevant checks.
        if perform_risk_audit:
            pkg_list_to_use = []
            project_type_to_use = project_type
            if project_type in risk_audit_map:
                pkg_list_to_use = pkg_list
            elif project_type in ("bom",):
                pkg_list_to_use = get_pkg_by_type(pkg_list, "npm")
                project_type_to_use = "npm"
            if pkg_list_to_use:
                if len(pkg_list) > 1:
                    console.print(
                        Panel(
                            f"Performing OSS Risk Audit for packages from "
                            f"{src_dir}\nNo of packages [bold]{len(pkg_list)}"
                            f"[/bold]. This will take a while ...",
                            title="OSS Risk Audit",
                            expand=False,
                        )
                    )
                try:
                    risk_results = risk_audit(
                        project_type_to_use,
                        scoped_pkgs,
                        args.private_ns,
                        pkg_list_to_use,
                    )
                    rtable, report_data = pkg_risks_table(
                        project_type_to_use,
                        scoped_pkgs,
                        risk_results,
                        pkg_max_risk_score=pkg_max_risk_score,
                        risk_report_file=risk_report_file,
                    )
                    if not args.no_vuln_table and report_data and rtable:
                        console.print(rtable)
                except Exception as e:
                    LOG.error(e)
                    LOG.error("Risk audit was not successful")
        # Do we support remote audit for this type?
        # Remote audits can improve results for some project types like npm by fetching vulnerabilities that might not yet be in our database.
        # In v6, remote audit is disabled by default and gets enabled with risk audit
        #
        # NOTE: Enabling risk audit may lead to some precision loss in reachability results.
        #   This is a known limitation with no immediate plan for resolution.
        if perform_risk_audit and project_type in type_audit_map:
            LOG.debug(
                "Performing remote audit for %s of type %s",
                src_dir,
                project_type,
            )
            LOG.debug("No of packages %d", len(pkg_list))
            try:
                audit_results = audit(project_type, pkg_list)
                if audit_results:
                    LOG.debug("Remote audit yielded %d results", len(audit_results))
                    results = results + audit_results
            except Exception as e:
                LOG.error("Remote audit was not successful")
                LOG.error(e)
                results = []
        # In case of docker, bom, or universal type, check if there are any
        # npm packages that can be audited remotely
        if perform_risk_audit and project_type in (
            "podman",
            "docker",
            "oci",
            "container",
            "bom",
            "universal",
        ):
            npm_pkg_list = get_pkg_by_type(pkg_list, "npm")
            if npm_pkg_list:
                LOG.debug("No of npm packages %d", len(npm_pkg_list))
                try:
                    audit_results = audit("nodejs", npm_pkg_list)
                    if audit_results:
                        LOG.debug(
                            "Remote audit yielded %d results",
                            len(audit_results),
                        )
                        results = results + audit_results
                except Exception as e:
                    LOG.error("Remote audit was not successful")
                    LOG.error(e)
        else:
            LOG.debug("Vulnerability database loaded from %s", config.VDB_BIN_FILE)
        if len(pkg_list) > 1:
            if project_type == "bom":
                LOG.info("Scanning CycloneDX xBOMs and atom slices")
            elif args.bom:
                LOG.info(
                    "Scanning %s with type %s",
                    args.bom,
                    project_type,
                )
            else:
                LOG.info(
                    "Scanning %s with type %s",
                    src_dir,
                    project_type,
                )
        # We could be dealing with multiple bom files
        bom_files = (
            get_all_bom_files(args.bom_dir)
            if args.bom_dir
            else [bom_file]
            if bom_file
            else []
        )
        if not pkg_list and not bom_files:
            LOG.debug("Empty package search attempted!")
        elif bom_files:
            LOG.debug("Scanning %d bom files for issues", len(bom_files))
        else:
            LOG.debug("Scanning %d oss dependencies for issues", len(pkg_list))
        # There are many ways to perform reachability analysis.
        # Most tools—including commercial ones—rely on a vulnerability database with affected modules (sinks) to detect reachable flows.
        # This has several downsides:
        # 1. These databases are often incomplete and manually maintained.
        # 2. If a CVE or ADP enhancement isn’t available yet, reachability won’t be detected.
        #
        # In contrast, depscan computes reachable flows (via atom) without relying on vulnerability data upfront.
        # It then identifies a smaller subset of those flows that are actually vulnerable.
        # From there, we can further narrow it down to flows that are Endpoint-Reachable, Exploitable, Container-Escapable, etc.
        reachability_analyzer = depscan_options.get("reachability_analyzer")
        reachability_options = None
        if (
            reachability_analyzer and reachability_analyzer != "off"
        ) or depscan_options.get("profile") != "generic":
            reachability_options = ReachabilityAnalysisKV(
                project_types=[project_type],
                src_dir=src_dir,
                bom_dir=args.bom_dir or reports_dir,
                require_multi_usage=depscan_options.get("require_multi_usage", False),
                source_tags=depscan_options.get("source_tags"),
                sink_tags=depscan_options.get("sink_tags"),
            )
        # Let’s proceed with the VDR analysis.
        summary, vdr_file, vdr_result = vdr_analyze_summarize(
            project_type,
            results,
            suggest_mode=args.suggest,
            scoped_pkgs=scoped_pkgs,
            bom_file=bom_files[0] if len(bom_files) == 1 else None,
            bom_dir=args.bom_dir,
            reports_dir=args.reports_dir,
            pkg_list=pkg_list,
            reachability_analyzer=reachability_analyzer,
            reachability_options=reachability_options,
            no_vuln_table=args.no_vuln_table,
            fuzzy_search=depscan_options.get("fuzzy_search", False),
            search_order=depscan_options.get("search_order"),
        )
        if vdr_result.pkg_vulnerabilities:
            all_pkg_vulnerabilities += vdr_result.pkg_vulnerabilities
        if vdr_result.prioritized_pkg_vuln_trees:
            all_pkg_group_rows.update(vdr_result.prioritized_pkg_vuln_trees)
        # Explain the results
        if args.explain:
            explainer.explain(
                project_type,
                src_dir,
                args.bom_dir or reports_dir,
                vdr_file,
                vdr_result,
                args.explanation_mode,
                depscan_options["llm_prompt_file"],
            )
        else:
            LOG.debug(
                "Pass the `--explain` argument to get a detailed explanation of the analysis."
            )
        # CSAF VEX export
        if args.csaf:
            export_csaf(
                vdr_result,
                src_dir,
                reports_dir,
                vdr_file,
            )
    console.record = True
    # Export the console output
    console.save_html(
        html_report_file,
        clear=False,
        theme=(MONOKAI if os.getenv("USE_DARK_THEME") else DEFAULT_TERMINAL_THEME),
    )
    console.save_text(txt_report_file, clear=False)
    # Prettify the rich html report
    html_report_generator = ReportGenerator(
        input_rich_html_path=html_report_file,
        report_output_path=html_report_file,
        raw_content=False,
    )
    html_report_generator.parse_and_generate_report()
    # This logic needs refactoring
    # render report into template if wished
    if args.report_template and os.path.isfile(args.report_template):
        utils.render_template_report(
            vdr_file=vdr_file,
            bom_file=bom_file,
            pkg_vulnerabilities=all_pkg_vulnerabilities,
            pkg_group_rows=all_pkg_group_rows,
            summary=summary,
            template_file=args.report_template,
            result_file=os.path.join(reports_dir, args.report_name),
            depscan_options=depscan_options,
        )
    elif args.report_template:
        LOG.warning(
            "Template file %s doesn't exist, custom report not created.",
            args.report_template,
        )
    # Should we include the generated text report as an annotation in the VDR file?
    if args.explain or args.annotate:
        annotate_vdr(vdr_file, txt_report_file)


def main():
    cli_args = build_args()
    run_depscan(cli_args)


if __name__ == "__main__":
    main()



import os
import shutil
import sys
import uuid
from collections import defaultdict
from datetime import datetime, timezone
from urllib.parse import unquote_plus

from custom_json_diff.lib.utils import json_load, json_dump
from defusedxml.ElementTree import parse
from xbom_lib.blint import BlintGenerator
from xbom_lib.cdxgen import (
    CdxgenGenerator,
    CdxgenImageBasedGenerator,
    CdxgenServerGenerator,
)
from depscan.lib.logger import LOG, SPINNER, console
from depscan.lib.utils import cleanup_license_string
from typing import Dict, Optional


def parse_bom_ref(bomstr, licenses=None):
    """
    Method to parse bom ref string into individual constituents

    :param bomstr: Bom ref string
    :param licenses: Licenses
    :return Dict containing group, name, and version for the package
    """
    if bomstr:
        bomstr = unquote_plus(bomstr)
    tmpl = bomstr.split("/")
    vendor = ""
    name_ver = []
    if len(tmpl) == 2:
        # Just name and version
        vendor = tmpl[0]
        name_ver = tmpl[1].split("@")
    elif len(tmpl) == 3:
        vendor = tmpl[1]
        name_ver = tmpl[-1].split("@")
    elif len(tmpl) > 3:
        vendor = tmpl[-2]
        name_ver = tmpl[-1].split("@")
    vendor = vendor.replace("pkg:", "")
    # If name starts with @ this will make sure the name still gets captured
    if len(name_ver) >= 2:
        name = name_ver[-2]
        version = name_ver[-1]
    else:
        name = name_ver[0]
        version = "*"
    if "?" in version:
        version = version.split("?")[0]
    if version.startswith("v"):
        version = version[1:]
    return {
        "vendor": vendor,
        "name": name,
        "version": version,
        "licenses": licenses,
    }


def get_licenses(ele):
    """
    Retrieve licenses from xml

    :param ele: An XML element
    :return A list of extracted licenses
    """
    license_list = []
    namespace = "{http://cyclonedx.org/schema/bom/1.5}"
    for data in ele.findall(f"{namespace}licenses/{namespace}license/{namespace}id"):
        license_list.append(data.text)
    if not license_list:
        for data in ele.findall(
            f"{namespace}licenses/{namespace}license/{namespace}name"
        ):
            if data is not None and data.text:
                ld_list = [data.text]
                if "http" in data.text:
                    ld_list = [
                        os.path.basename(data.text)
                        .replace(".txt", "")
                        .replace(".html", "")
                    ]
                elif "/" in data.text:
                    ld_list = [cleanup_license_string(data.text)]
                for ld in ld_list:
                    license_list.append(ld.strip().upper())
    return license_list


def get_package(component_ele, licenses):
    """
    Retrieve package from xml

    :param component_ele: The XML element representing a component.
    :param licenses: A list of licenses associated with the component.
    :return: A dictionary containing the package information
    """
    bom_ref = component_ele.attrib.get("bom-ref")
    pkg = {
        "licenses": licenses,
        "vendor": "",
        "name": "",
        "version": "",
        "scope": "",
    }
    if bom_ref and "/" in bom_ref:
        pkg = parse_bom_ref(bom_ref, licenses)
    for ele in component_ele.iter():
        if ele.tag.endswith("group") and ele.text:
            pkg["vendor"] = ele.text
        if ele.tag.endswith("name") and ele.text and not pkg["name"]:
            pkg["name"] = ele.text
        if ele.tag.endswith("version") and ele.text:
            version = ele.text
            if version.startswith("v"):
                version = version[1:]
            pkg["version"] = version
        if ele.tag.endswith("purl") and ele.text and not pkg.get("vendor"):
            purl = ele.text
            namespace = purl.split("/")[0].replace("pkg:", "")
            pkg["vendor"] = namespace
    return pkg


def get_pkg_list_json(jsonfile):
    """
    Method to extract packages from a bom json file

    :param jsonfile: Path to a bom json file.
    return List of dicts representing extracted packages
    """
    pkgs = []
    if bom_data := json_load(jsonfile, log=LOG):
        if bom_data.get("components"):
            for comp in bom_data.get("components", []):
                licenses, vendor, url = get_license_vendor_url(comp)
                pkgs.append(
                    {**comp, "vendor": vendor, "licenses": licenses, "url": url}
                )
        return pkgs


def get_license_vendor_url(comp):
    licenses = []
    vendor = comp.get("group") or ""
    if comp.get("licenses"):
        for lic in comp.get("licenses"):
            license_obj = lic
            if lic.get("license"):
                license_obj = lic.get("license")
            if license_obj.get("id"):
                licenses.append(license_obj.get("id"))
            elif license_obj.get("name"):
                licenses.append(cleanup_license_string(license_obj.get("name")))
    url = ""
    for aref in comp.get("externalReferences", []):
        if aref.get("type") in (
            "vcs",
            "issue-tracker",
            "website",
            "bom",
            "source-distribution",
            "distribution",
            "distribution-intake",
            "build-system",
            "model-card",
            "evidence",
            "formulation",
        ):
            url = aref.get("url", "")
            break
    return licenses, vendor, url


# Unused
def get_pkg_list(xmlfile):
    """Method to parse the bom xml file and convert into packages list

    :param xmlfile: BOM xml file to parse
    :return list of package dict
    """
    if xmlfile.endswith(".json"):
        return get_pkg_list_json(xmlfile)
    pkgs = []
    try:
        et = parse(xmlfile)
        root = et.getroot()
        for child in root:
            if child.tag.endswith("components"):
                for ele in child.iter():
                    if ele.tag.endswith("component"):
                        licenses = get_licenses(ele)
                        pkgs.append(get_package(ele, licenses))
    except Exception as pe:
        LOG.debug("Unable to parse %s %s", xmlfile, pe)
        LOG.warning(
            "Unable to produce Software Bill-of-Materials for this project. "
            "Execute the scan after installing the dependencies!"
        )
    return pkgs


def get_pkg_by_type(pkg_list, pkg_type):
    """Method to filter packages based on package type

    :param pkg_list: List of packages
    :param pkg_type: Package type to filter
    :return List of packages matching pkg_type
    """
    if not pkg_list:
        return []
    return [
        pkg for pkg in pkg_list if pkg.get("purl", "").startswith("pkg:" + pkg_type)
    ]


def create_bom(bom_file, src_dir=".", options=None):
    """
    Method to create BOM file by executing cdxgen command

    :param bom_file: BOM file
    :param src_dir: Source directory
    :param options: Additional options for generating the BOM file.
    :returns: True if the command was executed. False if the executable was
    not found.
    """
    if not options:
        options = {}
    # Get the various options and filenames
    techniques = options.get("techniques") or []
    lifecycles = options.get("lifecycles") or []
    project_type_list = options.get("project_type") or []
    bom_engine = options.get("bom_engine", "")
    lifecycle_analysis_mode = options.get("lifecycle_analysis_mode", False)
    # Detect if blint needs to be used for the given project type, technique, and lifecycle.
    # For binaries, generate an sbom with blint directly
    if (
        bom_engine == "BlintGenerator"
        or "binary-analysis" in techniques
        or "post-build" in lifecycles
        or any([t in ("binary", "apk") for t in project_type_list])
    ):
        return create_blint_bom(bom_file, src_dir, options=options)
    cdxgen_server = options.get("cdxgen_server")
    cdxgen_lib = CdxgenGenerator

    # Should we call cdxgen server
    if cdxgen_server or bom_engine == "CdxgenServerGenerator":
        if not cdxgen_server:
            LOG.error(
                "Pass the `--cdxgen-server` argument to use the cdxgen server for BOM generation. Alternatively, use `--bom-engine auto` or `--bom-engine CdxgenGenerator`."
            )
            return False
        cdxgen_lib = CdxgenServerGenerator
    else:
        # Prefer the new image based generators if docker command is available in auto mode
        if bom_engine == "CdxgenImageBasedGenerator":
            cdxgen_lib = CdxgenImageBasedGenerator
        elif bom_engine == "auto":
            # Prefer local CLI while scanning container images
            if any(
                [
                    t in ("docker", "podman", "oci", "os", "hardware")
                    for t in project_type_list
                ]
            ):
                cdxgen_lib = CdxgenGenerator
                if lifecycle_analysis_mode:
                    LOG.warning(
                        "Lifecycle analysis is not supported for oci and os project types."
                    )
                    lifecycle_analysis_mode = True
            elif (
                shutil.which(os.getenv("DOCKER_CMD", "docker"))
                and sys.platform != "win32"
            ):
                cdxgen_lib = CdxgenImageBasedGenerator
    # We now have the cdxgen library to use.
    # For lifecycle analysis, we need to generate multiple BOM files
    if lifecycle_analysis_mode:
        return create_lifecycle_boms(cdxgen_lib, src_dir, options)
    # Invoke the cdxgen library directly
    with console.status(
        f"Generating BOM for the source '{src_dir}' with cdxgen.", spinner=SPINNER
    ):
        bom_result = cdxgen_lib(
            src_dir, bom_file, logger=LOG, options=options
        ).generate()
        if not bom_result.success:
            LOG.info(
                "The cdxgen invocation was unsuccessful. Try generating the BOM separately."
            )
            LOG.debug(bom_result.command_output)
        return bom_result.success and os.path.exists(bom_file)


def create_blint_bom(
    bom_file: str, src_dir: str = ".", options: Optional[Dict] = None
) -> bool:
    """
    Method to create BOM file by using blint

    :param bom_file: BOM file
    :param src_dir: Source directory
    :param options: Additional options for generating the BOM file.
    :returns: True if the bom was generated successfully. False otherwise.
    """
    if options is None:
        options = {}
    reachability_analyzer = options.get("reachability_analyzer")
    # The side effect is that we will almost always run blint in deep mode
    if reachability_analyzer != "off" and not options.get("deep"):
        options["deep"] = True
    blint_lib = BlintGenerator(src_dir, bom_file, logger=LOG, options=options)
    with console.status(
        f"Generating BOM for the source '{src_dir}' with blint.", spinner=SPINNER
    ):
        bom_result = blint_lib.generate()
        if not bom_result.success:
            LOG.info(
                "The blint invocation was unsuccessful. Try generating the BOM separately."
            )
        return bom_result.success and os.path.exists(bom_file)


def create_lifecycle_boms(cdxgen_lib, src_dir, options):
    """
    Method to create multiple BOM files for each lifecycle

    :param cdxgen_lib: cdxgen library to use
    :param src_dir: Source directory
    :param options: Additional options for generating the BOM files
    """
    lifecycles = options.get("lifecycles", []) or []
    if lifecycles:
        LOG.warning(
            "Ignoring the `lifecycles` argument, as it is not required for lifecycle analysis."
        )
    any_success = False
    prebuild_bom_file = options.get("prebuild_bom_file")
    build_bom_file = options.get("build_bom_file")
    postbuild_bom_file = options.get("postbuild_bom_file")
    container_bom_file = options.get("container_bom_file")
    reachability_analyzer = options.get("reachability_analyzer")
    with console.status(
        f"Generating lifecycle-specific BOMs for {src_dir}.", spinner=SPINNER
    ) as status:
        # Start with build BOM generation.
        # This would help atom compute reachable slices from a build perspective without getting confused
        # about the pre-build state.
        status.update(f"Generating build BOM for '{src_dir}' with cdxgen.")
        coptions = {**options, "deep": "true", "lifecycles": ["build"]}
        # We must also run it under research profile to help the reachability analyzer
        # This logic could get refactored in the future
        if reachability_analyzer != "off" and options.get("profile") != "research":
            coptions["profile"] = "research"
        bom_result = cdxgen_lib(
            src_dir, build_bom_file, logger=LOG, options=coptions
        ).generate()
        if not bom_result.success or not os.path.exists(build_bom_file):
            LOG.debug(
                "The cdxgen invocation was unsuccessful. Trying pre-build lifecycle."
            )
            LOG.debug(bom_result.command_output)
        else:
            any_success = True
        # pre-build
        status.update(f"Now generating pre-build BOM for '{src_dir}' with cdxgen.")
        coptions = {**options, "deep": "false", "lifecycles": ["pre-build"]}
        bom_result = cdxgen_lib(
            src_dir, prebuild_bom_file, logger=LOG, options=coptions
        ).generate()
        if not bom_result.success or not os.path.exists(prebuild_bom_file):
            LOG.debug(
                "The cdxgen invocation was unsuccessful. Trying the build lifecycle."
            )
            LOG.debug(bom_result.command_output)
        else:
            any_success = True
        # container bom. For this we need the image name.
        container_image_name = os.getenv("DEPSCAN_SOURCE_IMAGE") or options.get(
            "source_image"
        )
        if container_image_name:
            status.update(f"Generating container BOM for '{src_dir}' with cdxgen.")
            coptions = {**options, "deep": "true", "project_type": ["oci"]}
            if container_image_name == src_dir:
                LOG.info(
                    "Set the environment variable DEPSCAN_SOURCE_IMAGE to the name of the container image to include its components."
                )
            bom_result = cdxgen_lib(
                container_image_name, container_bom_file, logger=LOG, options=coptions
            ).generate()
            if not bom_result.success or not os.path.exists(container_bom_file):
                LOG.debug(
                    "The cdxgen invocation was unsuccessful. Trying for the next lifecycle."
                )
                LOG.debug(bom_result.command_output)
            else:
                any_success = True
        else:
            LOG.debug(
                "Set the environment variable DEPSCAN_SOURCE_IMAGE to the name of the container image to include its components."
            )
        status.update("Preparing blint for post-build BOM generation.")
    # post-build BOM with blint
    coptions = {
        **options,
        "deep": False,
        "use_blintdb": False,
        "lifecycles": ["post-build"],
    }
    # What if the build directory is different to the source
    build_dir = os.getenv("DEPSCAN_BUILD_DIR") or options.get("build_dir") or src_dir
    res = create_blint_bom(postbuild_bom_file, build_dir, options=coptions)
    if not res or not os.path.exists(postbuild_bom_file):
        LOG.debug(
            "The blint invocation was unsuccessful. Try building this project prior to invoking depscan. Alternatively, check if this project generates binary artefacts."
        )
    else:
        any_success = True
    return any_success


def create_empty_vdr(pkg_list, ds_version):
    components = pkg_list or []
    bom_data = update_tools_metadata(None, None, ds_version)
    return {**bom_data, "components": components}


def update_tools_metadata(tools, bom_data, ds_version):
    """
    Helper function to add depscan information as metadata
    :param tools: Tools section of the SBOM
    :param bom_data: SBOM data
    :param ds_version: depscan version
    :return: None
    """
    if not bom_data:
        now_utc = datetime.now(timezone.utc)
        bom_data = {
            "bomFormat": "CycloneDX",
            "specVersion": "1.6",
            "serialNumber": f"urn:uuid:{uuid.uuid4()}",
            "version": 1,
            "metadata": {
                "timestamp": now_utc.strftime("%Y-%m-%dT%H:%M:%SZ"),
            },
        }
    components = tools.get("components", []) if tools else []
    needs_ds_component = (
        len([c for c in components if c.get("name") == "owasp-depscan"]) == 0
    )
    if needs_ds_component:
        ds_purl = f"pkg:pypi/owasp-depscan@{ds_version}"
        components.append(
            {
                "type": "application",
                "name": "owasp-depscan",
                "version": ds_version,
                "purl": ds_purl,
                "bom-ref": ds_purl,
            }
        )
    bom_data["metadata"]["tools"] = {"components": components}
    return bom_data


def export_bom(bom_data, ds_version, pkg_vulnerabilities, vdr_file):
    """
    Exports the Bill of Materials (BOM) data along with package vulnerabilities
    to a Vulnerability Data Report (VDR) file.

    :param bom_data: SBOM data
    :param ds_version: depscan version
    :param pkg_vulnerabilities: Package vulnerabilities
    :param vdr_file: VDR file path
    """
    # Add depscan information as metadata
    metadata = bom_data.get("metadata", {})
    tools = metadata.get("tools", {})
    bom_version = str(bom_data.get("version", 0))
    # Update the version
    if bom_version.isdigit():
        bom_data["version"] = int(bom_version) + 1
    # Update the tools section
    if isinstance(tools, dict):
        bom_data = update_tools_metadata(tools, bom_data, ds_version)
    bom_data = trim_vdr_bom_data(bom_data)
    bom_data["vulnerabilities"] = pkg_vulnerabilities
    json_dump(
        vdr_file,
        bom_data,
        compact=True,
        error_msg=f"Unable to generate VDR file at {vdr_file}",
    )


def trim_vdr_bom_data(bom_data):
    components = bom_data.get("components")
    if not components:
        return bom_data
    metadata = bom_data.get("metadata")
    if metadata and metadata.get("properties"):
        del metadata["properties"]
        bom_data["metadata"] = metadata
    new_components = {}
    component_identities = defaultdict(list)
    for comp in components:
        identity_evidences = comp.get("evidence", {}).get("identity", []) or []
        if isinstance(identity_evidences, dict):
            identity_evidences = [identity_evidences]
        for p in (
            "properties",
            "signature",
            "url",
            "vendor",
            "licenses",  # We need a better logic to retain licenses here
        ):
            if comp.get(p) is not None:
                del comp[p]
        ref = comp.get("bom-ref") or comp.get("purl")
        # This is an error condition really
        if not ref:
            continue
        component_identities[ref] += identity_evidences
        if not new_components.get(ref):
            new_components[ref] = comp
    vdr_components = []
    for ref, comp in new_components.items():
        identity_evidences = component_identities[ref]
        comp["evidence"] = {"identity": identity_evidences}
        vdr_components.append(comp)
    bom_data["components"] = vdr_components
    for p in (
        "annotations",
        "signature",
    ):
        if bom_data.get(p):
            del bom_data[p]
    return bom_data


def annotate_vdr(vdr_file, txt_report_file):
    if (
        not vdr_file
        or not txt_report_file
        or not os.path.exists(vdr_file)
        or not os.path.exists(txt_report_file)
    ):
        return
    vdr = json_load(vdr_file)
    metadata = vdr.get("metadata", {})
    # Some cyclonedx sbom don't containg tools.components
    if "components" in metadata.get("tools"):
        tools = metadata.get("tools", {}).get("components", {})
    else:
        tools = {}
    with open(txt_report_file, errors="ignore", encoding="utf-8") as txt_fp:
        report = txt_fp.read()
        annotations = vdr.get("annotations", []) or []
        depscan_annotation = {
            "subjects": [vdr.get("serialNumber")],
            "annotator": {"component": tools[-1] if len(tools) > 0 else {}},
            "timestamp": metadata.get("timestamp"),
            "text": report,
        }
        annotations.append(depscan_annotation)
    vdr["annotations"] = annotations
    json_dump(
        vdr_file,
        vdr,
        compact=True,
        error_msg=f"Unable to add annotations to the VDR file at {vdr_file}",
    )



from datetime import datetime

from depscan.lib import config
from semver import Version
from depscan.lib.package_query.pkg_query import compute_time_risks, calculate_risk_score


def pypi_pkg_risk(pkg_metadata, is_private_pkg, scope, pkg):
    """
    Calculate various package risks based on the metadata from pypi.

    :param pkg_metadata: A dict containing the metadata of the package from PyPI
    :param is_private_pkg: Boolean to indicate if this package is private
    :param scope: Package scope
    :param pkg: Package object

    :return: Dict of risk metrics and corresponding PyPI values.
    """
    risk_metrics = {
        "pkg_deprecated_risk": False,
        "pkg_version_deprecated_risk": False,
        "pkg_version_missing_risk": False,
        "pkg_min_versions_risk": False,
        "created_now_quarantine_seconds_risk": False,
        "latest_now_max_seconds_risk": False,
        "mod_create_min_seconds_risk": False,
        "pkg_min_maintainers_risk": False,
        "pkg_private_on_public_registry_risk": False,
    }
    info = pkg_metadata.get("info", {})
    versions_dict = pkg_metadata.get("releases", {})
    versions = [ver[0] for k, ver in versions_dict.items() if ver]
    is_deprecated = info.get("yanked") and info.get("yanked_reason")
    is_version_deprecated = False
    if not is_deprecated and pkg and pkg.get("version"):
        theversion = versions_dict.get(pkg.get("version"), [])
        if isinstance(theversion, list) and len(theversion) > 0:
            theversion = theversion[0]
        elif theversion and theversion.get("yanked"):
            is_version_deprecated = True
        # Check if the version exists in the registry
        if not theversion:
            risk_metrics["pkg_version_missing_risk"] = True
            risk_metrics["pkg_version_missing_value"] = 1
    # Some packages like pypi:azure only mention deprecated in the description
    # without yanking the package
    pkg_description = info.get("description", "").lower()
    if not is_deprecated and (
        "is deprecated" in pkg_description
        or "no longer maintained" in pkg_description
    ):
        is_deprecated = True
    latest_deprecated = False
    version_nums = list(versions_dict.keys())
    # Ignore empty versions without metadata. Thanks pypi
    version_nums = [ver for ver in version_nums if versions_dict.get(ver)]
    try:
        first_version_num = min(
            version_nums,
            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
        )
        latest_version_num = max(
            version_nums,
            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
        )
    except (ValueError, TypeError):
        first_version_num = version_nums[0]
        latest_version_num = version_nums[-1]
    first_version = versions_dict.get(first_version_num)[0]
    latest_version = versions_dict.get(latest_version_num)[0]

    # Is the private package available publicly? Dependency confusion.
    if is_private_pkg and pkg_metadata:
        risk_metrics["pkg_private_on_public_registry_risk"] = True
        risk_metrics["pkg_private_on_public_registry_value"] = 1

    # If the package has fewer than minimum number of versions
    if len(versions):
        if len(versions) < config.pkg_min_versions:
            risk_metrics["pkg_min_versions_risk"] = True
            risk_metrics["pkg_min_versions_value"] = len(versions)
        # Check if the latest version is deprecated
        if latest_version and latest_version.get("yanked"):
            latest_deprecated = True

    # Created and modified time related checks
    if first_version and latest_version:
        created = first_version.get("upload_time")
        modified = latest_version.get("upload_time")
        if created and modified:
            modified_dt = datetime.fromisoformat(modified)
            created_dt = datetime.fromisoformat(created)
            mod_create_diff = modified_dt - created_dt
            latest_now_diff = datetime.now() - modified_dt
            created_now_diff = datetime.now() - created_dt
            risk_metrics = compute_time_risks(
                risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
            )

    # Is the package deprecated
    if is_deprecated or latest_deprecated:
        risk_metrics["pkg_deprecated_risk"] = True
        risk_metrics["pkg_deprecated_value"] = 1
    elif is_version_deprecated:
        risk_metrics["pkg_version_deprecated_risk"] = True
        risk_metrics["pkg_version_deprecated_value"] = 1
    # Add package scope related weight
    if scope:
        risk_metrics[f"pkg_{scope}_scope_risk"] = True
        risk_metrics[f"pkg_{scope}_scope_value"] = 1

    risk_metrics["risk_score"] = calculate_risk_score(risk_metrics)
    return risk_metrics



import os
import re
import glob
from collections import defaultdict
from custom_json_diff.lib.utils import json_load
from rich import box
from rich.markdown import Markdown
from rich.table import Table
from rich.tree import Tree
from rich.panel import Panel

from depscan.lib.config import (
    COMMON_CHECK_TAGS,
    max_purl_per_flow,
    max_reachable_explanations,
    max_purls_reachable_explanations,
    max_source_reachable_explanations,
    max_sink_reachable_explanations,
    max_flows_per_prompt,
)
from depscan.lib.logger import console, LOG

SEVERITY_COLORS = {
    "CRITICAL": "bold red",
    "HIGH": "red",
    "MEDIUM": "yellow",
    "LOW": "blue",
    "INFO": "white",
    "UNKNOWN": "white",
}


def explain(
    project_type,
    src_dir,
    bom_dir,
    vdr_file,
    vdr_result,
    explanation_mode,
    prompts_file=None,
):
    """
    Explain the analysis and findings based on the explanation mode.

    :param project_type: Project type
    :param src_dir: Source directory
    :param bom_dir: BOM directory
    :param vdr_file: VDR file
    :param vdr_result: VDR Result
    :param explanation_mode: Explanation mode
    :param prompts_file: File path to save LLM prompts
    """
    pattern_methods = {}
    has_any_explanation = False
    has_any_crypto_flows = False
    slices_files = glob.glob(f"{bom_dir}/**/*reachables.slices*.json", recursive=True)
    openapi_spec_files = None

    purl_vuln_map = defaultdict(list)
    if vdr_result and vdr_result.pkg_vulnerabilities:
        for vuln in vdr_result.pkg_vulnerabilities:
            ref = vuln.get("bom-ref") or vuln.get("purl")
            if ref:
                purl_vuln_map[ref].append(vuln)

    # Should we explain the endpoints and Code Hotspots
    if explanation_mode in ("Endpoints", "EndpointsAndReachables", "LLMPrompts"):
        openapi_spec_files = glob.glob(f"{bom_dir}/*openapi*.json", recursive=False)
        if not openapi_spec_files:
            openapi_spec_files = glob.glob(f"{src_dir}/*openapi*.json", recursive=False)
    if openapi_spec_files:
        rsection = Markdown("""## Service Endpoints

The following endpoints and code hotspots were identified by depscan. Verify that proper authentication and authorization mechanisms are in place to secure them.""")
        any_endpoints_shown = False
        for ospec in openapi_spec_files:
            pattern_methods = print_endpoints(
                ospec, rsection if not any_endpoints_shown else None
            )
            if not any_endpoints_shown and pattern_methods:
                any_endpoints_shown = True

    # Return early for endpoints only explanations
    if explanation_mode in ("Endpoints",):
        return
    section_title = (
        "Non-Reachable Flows"
        if explanation_mode in ("NonReachables",)
        else "Reachable Flows"
    )
    for sf in slices_files:
        if len(slices_files) > 1:
            fn = os.path.basename(sf)
            section_label = f"# Explanations for {fn}"
            console.print(Markdown(section_label))
        if reachables_data := json_load(sf, log=LOG):
            # Backwards compatibility
            if isinstance(reachables_data, dict) and reachables_data.get("reachables"):
                reachables_data = reachables_data.get("reachables")
            if explanation_mode in ("NonReachables",):
                rsection = Markdown(
                    f"""## {section_title}

Below are several data flows deemed safe and non-reachable. Use the provided tips to confirm this assessment.
                """
                )
            elif pattern_methods:
                rsection = Markdown(
                    f"""## {section_title}

Below are some reachable flows, including those accessible via endpoints, identified by depscan. Use the generated OpenAPI specification to evaluate these endpoints for vulnerabilities and risk.
                """
                )
            else:
                rsection = Markdown(
                    f"""## {section_title}

Below are several data flows identified by depscan, including reachable ones. Use the tips provided to strengthen your application’s security posture.
                """
                )
            has_explanation, has_crypto_flows, tips = explain_reachables(
                explanation_mode,
                reachables_data,
                project_type,
                vdr_result,
                purl_vuln_map,
                prompts_file,
                rsection if not has_any_explanation else None,
            )
            if not has_any_explanation and has_explanation:
                has_any_explanation = True
            if not has_any_crypto_flows and has_crypto_flows:
                has_any_crypto_flows = True


def _track_usage_targets(usage_targets, usages_object):
    for k, v in usages_object.items():
        for file, lines in v.items():
            for l in lines:
                usage_targets.add(f"{file}#{l}")


def print_endpoints(ospec, header_section=None):
    if not ospec:
        return
    paths = json_load(ospec).get("paths") or {}
    pattern_methods = defaultdict(list)
    pattern_usage_targets = defaultdict(set)
    for pattern, path_obj in paths.items():
        usage_targets = set()
        http_method_added = False
        for k, v in path_obj.items():
            if k == "parameters":
                continue
            # Java, JavaScript, Python etc
            if k == "x-atom-usages":
                _track_usage_targets(usage_targets, v)
                continue
            if isinstance(v, dict) and v.get("x-atom-usages"):
                _track_usage_targets(usage_targets, v.get("x-atom-usages"))
            pattern_methods[pattern].append(k)
            http_method_added = True
        pattern_usage_targets[pattern] = usage_targets
        # We see an endpoint, but do not know the HTTP methods.
        # Let's track them as empty
        if not http_method_added and usage_targets:
            pattern_methods[pattern].append("")
    caption = ""
    if pattern_methods:
        caption = f"Identified Endpoints: {len(pattern_methods.keys())}"
    rtable = Table(
        box=box.DOUBLE_EDGE,
        show_lines=True,
        title="Endpoints",
        caption=caption,
    )
    for c in ("URL Pattern", "HTTP Methods", "Code Hotspots"):
        rtable.add_column(header=c, vertical="top")
    for k, v in pattern_methods.items():
        v.sort()
        sorted_areas = list(pattern_usage_targets[k])
        sorted_areas.sort()
        rtable.add_row(k, ("\n".join(v)).upper(), "\n".join(sorted_areas))
    if pattern_methods:
        # Print the header section
        if header_section:
            console.print(header_section)
        console.print()
        console.print(rtable)
    return pattern_methods


def is_cpp_flow(flows):
    if not flows:
        return False
    attempts = 0
    for idx, aflow in enumerate(flows):
        if aflow.get("parentFileName", "").endswith(".c") or aflow.get(
            "parentFileName", ""
        ).endswith(".cpp"):
            return True
        attempts += 1
        if attempts > 3:
            return False
    return False


def explain_reachables(
    explanation_mode,
    reachables,
    project_type,
    vdr_result,
    purl_vuln_map,
    prompts_file=None,
    header_section=None,
):
    """"""
    reachable_explanations = 0
    checked_flows = 0
    has_crypto_flows = False
    explained_ids = {}
    purls_reachable_explanations = defaultdict(int)
    source_reachable_explanations = defaultdict(int)
    sink_reachable_explanations = defaultdict(int)
    has_explanation = False
    header_shown = False
    has_cpp_flow = False
    purl_grouped_flows = defaultdict(list)

    # Backwards compatibility
    if isinstance(reachables, dict) and reachables.get("reachables"):
        reachables = reachables.get("reachables")
    for areach in reachables:
        cpp_flow = is_cpp_flow(areach.get("flows"))
        if not has_cpp_flow and cpp_flow:
            has_cpp_flow = True
        if (
            not areach.get("flows")
            or len(areach.get("flows")) < 2
            or (not areach.get("purls") and not cpp_flow)
        ):
            continue
        (
            flow_tree,
            added_ids,
            comment,
            source_sink_desc,
            source_code_str,
            sink_code_str,
            has_check_tag,
            is_endpoint_reachable,
            is_crypto_flow,
            max_severity,
            vuln_ids,
        ) = explain_flows(
            explanation_mode,
            areach.get("flows"),
            areach.get("purls"),
            project_type,
            vdr_result,
            purl_vuln_map,
        )
        # The goal is to reduce duplicate explanations by checking if a given flow is similar to one we have explained
        # before. We do this by checking the node ids, source-sink explanations, purl tags and so on.
        added_ids_str = "-".join(added_ids)
        # Have we seen this sequence before?
        if explained_ids.get(added_ids_str) or len(added_ids) < 4:
            continue
        if not source_sink_desc or not flow_tree or len(flow_tree.children) < 4:
            continue
        # In non-reachables mode, we are not interested in reachable flows.
        if (
            explanation_mode
            and explanation_mode in ("NonReachables",)
            and not has_check_tag
        ):
            continue
        if (
            source_code_str
            and source_reachable_explanations[source_code_str] + 1
            > max_source_reachable_explanations
        ):
            continue
        if (
            sink_code_str
            and sink_reachable_explanations[sink_code_str] + 1
            > max_sink_reachable_explanations
        ):
            continue
        purls_str = ",".join(sorted(areach.get("purls", [])))
        if (
            purls_str
            and purls_reachable_explanations[purls_str] + 1
            > max_purls_reachable_explanations
        ):
            continue
        if not has_explanation:
            has_explanation = True
        # Did we find any crypto flows
        if is_crypto_flow and not has_crypto_flows:
            has_crypto_flows = True

        title_style = "bold"
        if max_severity and max_severity in SEVERITY_COLORS:
            title_style = SEVERITY_COLORS[max_severity]
            source_sink_desc = f"[{title_style}][{max_severity} {', '.join(vuln_ids)}][/{title_style}] {source_sink_desc}"

        rtable = Table(
            box=box.DOUBLE_EDGE,
            show_lines=True,
            caption=comment,
            show_header=False,
            title=f"[bold]#{reachable_explanations + 1}[/bold] {source_sink_desc}",
            title_justify="left",
            min_width=150,
            border_style=title_style if max_severity else "white",
        )
        rtable.add_column(header="Flow", vertical="top")
        rtable.add_row(flow_tree)

        # Print the header first in case we haven't
        if not header_shown and header_section:
            console.print()
            console.print(header_section)
            header_shown = True
        console.print()
        console.print(rtable)

        if is_endpoint_reachable and explanation_mode in ("LLMPrompts",):
            for p in areach.get("purls", []):
                purl_grouped_flows[p].append(
                    {
                        "flows": areach.get("flows"),
                        "desc": source_sink_desc,
                        "id": reachable_explanations + 1,
                    }
                )

        explained_ids[added_ids_str] = True
        reachable_explanations += 1
        if purls_str:
            purls_reachable_explanations[purls_str] += 1
        if source_code_str:
            source_reachable_explanations[source_code_str] += 1
        if sink_code_str:
            sink_reachable_explanations[sink_code_str] += 1
        if has_check_tag:
            checked_flows += 1
        if reachable_explanations + 1 > max_reachable_explanations:
            break

    tips = """## Secure Design Tips"""
    if explanation_mode in ("NonReachables",):
        tips += """
- Automate tests (including fuzzing) to verify validation, sanitization, encoding, and encryption.
- Align the implementation with the original architecture and threat models to ensure security compliance.
- Extract reusable methods into a shared library for organization-wide use.
"""
    elif has_explanation:
        if has_crypto_flows:
            tips += """
- Generate a Cryptographic BOM with cdxgen and monitor it in Dependency-Track.
"""
        elif checked_flows:
            if not has_cpp_flow:
                tips += """
- Review the validation and sanitization methods used in the application.
- To enhance the security posture, implement a common validation middleware.
"""
            else:
                tips += """
- Continuously fuzz the parser and validation functions with diverse payloads.
- Generate post-build SBOMs with OWASP blint by building this project for various architecture combinations. Re-run depscan with the `--bom-dir` argument to enhance the analysis.
"""
        elif purls_reachable_explanations:
            tips += """
- Consider implementing a common validation and sanitization library to reduce the risk of exploitability.
"""
        else:
            tips += """
- Enhance your unit and integration tests to cover the flows listed above.
- Continuously fuzz the parser and validation functions with diverse payloads.
"""
    if tips and (has_explanation or has_crypto_flows):
        rsection = Markdown(tips)
        console.print(rsection)
    if purl_grouped_flows:
        save_llm_prompts(purl_grouped_flows, purl_vuln_map, project_type, prompts_file)
    return has_explanation, has_crypto_flows, tips


def get_vulns_for_purls(purls, purl_vuln_map):
    """
    Finds the highest severity vulnerability among the given purls.
    Returns (max_severity, [list_of_vuln_ids])
    """
    if not purls or not purl_vuln_map:
        return None, []

    found_vulns = []
    max_severity_val = 100
    severity_order = {
        "CRITICAL": 1,
        "HIGH": 2,
        "MEDIUM": 3,
        "LOW": 4,
        "INFO": 5,
        "UNKNOWN": 6,
    }

    current_max_severity = None

    for p in purls:
        vulns = purl_vuln_map.get(p, [])
        for v in vulns:
            v_id = v.get("id")
            v_sev = str(v.get("severity", "UNKNOWN")).upper()

            if v_id and v_id not in found_vulns:
                found_vulns.append(v_id)

            rank = severity_order.get(v_sev, 6)
            current_rank = severity_order.get(current_max_severity, 100)

            if rank < current_rank:
                current_max_severity = v_sev

    return current_max_severity, found_vulns


def flow_to_source_sink(idx, flow, purls, project_type, vdr_result, purl_vuln_map):
    """ """
    endpoint_reached_purls = {}
    reached_services = {}
    if vdr_result:
        endpoint_reached_purls = vdr_result.endpoint_reached_purls
        reached_services = vdr_result.reached_services
    is_endpoint_reachable = False
    possible_reachable_service = False
    tags = flow.get("tags", [])
    is_crypto_flow = "crypto" in tags or "crypto-generate" in tags
    method_in_emoji = ":right_arrow_curving_left:"
    for p in purls:
        if endpoint_reached_purls and endpoint_reached_purls.get(p):
            is_endpoint_reachable = True
            method_in_emoji = ":heavy_large_circle: "
        if reached_services and reached_services.get(p):
            possible_reachable_service = True
    max_severity, vuln_ids = get_vulns_for_purls(purls, purl_vuln_map)
    source_sink_desc = ""
    param_name = flow.get("name")
    method_str = "method"
    param_str = "Parameter"
    if param_name == "this":
        param_name = ""
    parent_file = flow.get("parentFileName", "")
    parent_method = flow.get("parentMethodName", "")
    # Improve the labels based on the language
    if re.search(".(js|ts|jsx|tsx|py|cs|php)$", parent_file):
        method_str = "function"
        param_str = "argument"
        if parent_method in ("handleRequest",):
            method_str = f"handler {method_str}"
        elif parent_method in ("__construct", "__init"):
            method_str = "constructor"
        elif project_type in ("php",) and parent_method.startswith("__"):
            method_str = f"magic {method_str}"
    if flow.get("label") == "METHOD_PARAMETER_IN":
        if param_name:
            source_sink_desc = f"""{param_str} [red]{param_name}[/red] {method_in_emoji} to the {method_str} [bold]{parent_method}[/bold]"""
        else:
            source_sink_desc = f"""{method_str.capitalize()} [red]{parent_method}[/red] {method_in_emoji}"""
    elif flow.get("label") == "CALL" and flow.get("isExternal"):
        method_full_name = flow.get("fullName", "")
        if not method_full_name.startswith("<"):
            source_sink_desc = f"Invocation: {method_full_name}"
    elif flow.get("label") == "RETURN" and flow.get("code"):
        source_sink_desc = flow.get("code").split("\n")[0]
    elif project_type not in ("java",) and flow.get("label") == "IDENTIFIER":
        source_sink_desc = flow.get("code").split("\n")[0]
        if source_sink_desc.endswith("("):
            source_sink_desc = f":diamond_suit: {source_sink_desc})"
    # Try to understand the source a bit more
    if source_sink_desc.startswith("require("):
        source_sink_desc = "The flow originates from a module import."
    elif (
        ".use(" in source_sink_desc
        or ".subscribe(" in source_sink_desc
        or ".on(" in source_sink_desc
        or ".emit(" in source_sink_desc
        or " => {" in source_sink_desc
    ):
        source_sink_desc = "The flow originates from a callback function."
    elif (
        "middleware" in source_sink_desc.lower() or "route" in source_sink_desc.lower()
    ):
        source_sink_desc = "The flow originates from middleware."
    elif len(purls) == 0:
        if tags:
            source_sink_desc = (
                f"{source_sink_desc} can be used to reach packages with tags `{tags}`"
            )
    elif len(purls) == 1:
        if is_endpoint_reachable:
            source_sink_desc = f"{source_sink_desc} can be used to reach this package from certain endpoints."
        elif source_sink_desc:
            if is_crypto_flow:
                source_sink_desc = "Reachable crypto-flow."
            else:
                source_sink_desc = "Reachable data-flow."
    else:
        if is_endpoint_reachable:
            source_sink_desc = f"{source_sink_desc} can be used to reach {len(purls)} packages from certain endpoints."
        else:
            if source_sink_desc:
                source_sink_desc = (
                    f"{source_sink_desc} can be used to reach {len(purls)} packages."
                )
            elif is_crypto_flow:
                source_sink_desc = (
                    f"{len(purls)} packages reachable from this crypto-flow."
                )
            else:
                source_sink_desc = (
                    f"{len(purls)} packages reachable from this data-flow."
                )
    return (
        source_sink_desc,
        is_endpoint_reachable,
        is_crypto_flow,
        max_severity,
        vuln_ids,
    )


def filter_tags(tags):
    if tags:
        tags = [
            atag
            for atag in tags.split(", ")
            if atag not in ("RESOLVED_MEMBER", "UNKNOWN_METHOD", "UNKNOWN_TYPE_DECL")
        ]
        return ", ".join(tags)
    return tags


def is_filterable_code(project_type, code):
    if len(code) < 3:
        return True
    for c in (
        "console.log",
        "thoughtLog(",
        "_tmp_",
        "LOG.debug(",
        "options.get(",
        "RET",
        "this.",
        "NULL",
        "!",
    ):
        if code and code.startswith(c):
            return True
    return False


def flow_to_str(explanation_mode, flow, project_type):
    """"""
    has_check_tag = False
    file_loc = ""
    if (
        flow.get("parentFileName")
        and flow.get("lineNumber")
        and not flow.get("parentFileName").startswith("unknown")
    ):
        # strip common prefixes
        name = flow.get("parentFileName", "")
        for p in ("src/main/java/", "src/main/scala/"):
            name = name.removeprefix(p)
        file_loc = f"{name}#{flow.get('lineNumber')}    "
    node_desc = flow.get("code").split("\n")[0]
    if (len(node_desc) < 3 or node_desc.endswith("{")) and len(flow.get("code")) > 3:
        node_desc = " ".join(flow.get("code", "").split())
        if "(" in node_desc:
            node_desc = node_desc.split("(")[0] + "() ..."
    if node_desc.endswith("("):
        node_desc = f":diamond_suit: {node_desc})"
    elif node_desc.startswith("return "):
        node_desc = f":arrow_backward: [italic]{node_desc}[/italic]"
    tags = filter_tags(flow.get("tags"))
    if flow.get("label") in ("METHOD_PARAMETER_IN",):
        param_name = flow.get("name")
        if param_name == "this":
            param_name = ""
        node_desc = f"{flow.get('parentMethodName')}([red]{param_name}[/red]) :right_arrow_curving_left:"
        if tags:
            node_desc = f"{node_desc}\n[bold]Tags:[/bold] [italic]{tags}[/italic]\n"
    elif flow.get("label") in ("IDENTIFIER", "CALL"):
        if node_desc.startswith("<"):
            node_desc = flow.get("name")
        if flow.get("isExternal"):
            node_desc = f"{node_desc} :right_arrow_curving_up:"
        if tags:
            node_desc = f"{node_desc}\n[bold]Tags:[/bold] [italic]{tags}[/italic]\n"
    if tags and not is_filterable_code(project_type, node_desc):
        for ctag in COMMON_CHECK_TAGS:
            if ctag in tags:
                has_check_tag = True
                break
    if has_check_tag:
        if explanation_mode in ("NonReachables",):
            node_desc = f"[bold][green]{node_desc}[/green][/bold]"
        else:
            node_desc = f"[green]{node_desc}[/green]"
    flow_str = (
        f"""[gray37]{file_loc}[/gray37]{node_desc}"""
        if not is_filterable_code(project_type, node_desc)
        else ""
    )
    return (
        file_loc,
        flow_str,
        node_desc,
        has_check_tag,
    )


def get_comment_symbol(filename):
    """
    Returns the appropriate comment symbol based on the file extension.
    """
    if filename.endswith((".py", ".rb", ".sh", ".pl", ".yaml", ".yml", ".dockerfile")):
        return "#"
    return "//"


def generate_grouped_llm_prompt(purl, flows_data, purl_vuln_map, project_type):
    """
    Generates a consolidated prompt for a specific PURL and its reachable flows.
    """
    max_sev, vulns = get_vulns_for_purls([purl], purl_vuln_map)
    vuln_text = f", associated with {', '.join(vulns)}" if vulns else ""

    prompt = f"## LLM Prompt for `{purl}`\n\n"
    prompt += f"Analyze the security risks for the package `{purl}`{vuln_text}. "
    prompt += f"depscan identified {len(flows_data)} endpoint reachable data flows leading to this package.\n\n"

    count = 0
    for item in flows_data:
        if count >= max_flows_per_prompt:
            prompt += f"... (and {len(flows_data) - max_flows_per_prompt} more flows)\n"
            break

        desc = item.get("desc", "Unknown flow")
        desc = re.sub(r"\[.*?]", "", desc)

        prompt += f"### Flow #{item.get('id')}: {desc}\n\n"
        last_code = ""
        for flow in item.get("flows", []):
            code = flow.get("code", "").strip()
            if last_code and last_code == code:
                continue
            if len(code) < 8:
                continue
            last_code = code
            filename = flow.get("parentFileName", "")
            lineno = flow.get("lineNumber", "")
            tags = flow.get("tags")
            if tags:
                comment_symbol = get_comment_symbol(filename)
                code = f"{comment_symbol} {tags}\n{code}"
            prompt += f"- Source: `{filename}:{lineno}`\n```\n{code}\n```\n"
        prompt += "\n"
        count += 1

    prompt += "Based on these data flows, explain how an attacker might exploit the vulnerabilities in this package and suggest secure code fixes or mitigations."
    return prompt


def save_llm_prompts(purl_grouped_flows, purl_vuln_map, project_type, prompts_file):
    """
    Iterates through grouped flows and saves consolidated LLM prompts to a file.
    """
    if not prompts_file:
        return
    with open(prompts_file, "w", encoding="utf-8") as f:
        f.write("# LLM Prompts generated by OWASP dep-scan\n\n")
        for purl, flows_data in purl_grouped_flows.items():
            prompt_text = generate_grouped_llm_prompt(
                purl, flows_data, purl_vuln_map, project_type
            )
            f.write(prompt_text)
            f.write("\n---\n\n")


def explain_flows(
    explanation_mode, flows, purls, project_type, vdr_result, purl_vuln_map
):
    """"""
    tree = None
    comments = []
    if len(purls) > max_purl_per_flow:
        comments.append(
            ":exclamation_mark: Refactor this flow to minimize the use of external libraries."
        )
    if purls:
        purls_display = []
        for p in purls:
            max_sev, vulns = get_vulns_for_purls([p], purl_vuln_map)
            if max_sev:
                style = SEVERITY_COLORS.get(max_sev, "white")
                purls_display.append(f"[{style}]{p} ({', '.join(vulns)})[/{style}]")
            else:
                purls_display.append(p)

        purls_str = "\n".join(purls_display)
        comments.append(f"[info]Reachable Packages:[/info]\n{purls_str}")

    added_ids = []
    added_flows = []
    added_node_desc = []
    has_check_tag = False
    last_file_loc = None
    source_sink_desc = ""
    last_code = ""
    source_code_str = ""
    sink_code_str = ""
    is_endpoint_reachable = False
    is_crypto_flow = False
    max_severity = None
    vuln_ids = []

    for idx, aflow in enumerate(flows):
        # For java, we are only interested in identifiers with tags to keep the flows simple to understand
        if (
            project_type in ("java", "jar", "android", "apk")
            and aflow.get("label") == "IDENTIFIER"
            and not aflow.get("tags")
        ):
            continue
        curr_code = aflow.get("code", "").split("\n")[0]
        if idx == 0:
            source_code_str = curr_code
        if idx == len(flows):
            sink_code_str = curr_code
        if last_code and last_code == curr_code:
            continue
        last_code = curr_code
        if not source_sink_desc:
            (
                source_sink_desc,
                is_endpoint_reachable,
                is_crypto_flow,
                max_severity,
                vuln_ids,
            ) = flow_to_source_sink(
                idx, aflow, purls, project_type, vdr_result, purl_vuln_map
            )
        file_loc, flow_str, node_desc, has_check_tag_flow = flow_to_str(
            explanation_mode, aflow, project_type
        )
        if not flow_str or (last_file_loc and last_file_loc == file_loc):
            continue
        last_file_loc = file_loc
        if flow_str in added_flows or node_desc in added_node_desc:
            continue
        added_flows.append(flow_str)
        added_ids.append(str(aflow.get("id", "")))
        added_node_desc.append(node_desc)
        if not tree:
            tree = Tree(flow_str)
        else:
            tree.add(flow_str)
        if has_check_tag_flow:
            has_check_tag = True
    if has_check_tag and explanation_mode not in ("NonReachables",):
        comments.insert(
            0,
            ":white_medium_small_square: Verify that the mitigation(s) used in this flow are valid and appropriate for your security requirements.",
        )
    return (
        tree,
        added_ids,
        "\n".join(comments),
        source_sink_desc,
        source_code_str,
        sink_code_str,
        has_check_tag,
        is_endpoint_reachable,
        is_crypto_flow,
        max_severity,
        vuln_ids,
    )



from datetime import datetime

from depscan.lib import config
from depscan.lib.package_query.pkg_query import httpclient, compute_time_risks, calculate_risk_score

def search_npm(keywords=None, insecure_only=False, unstable_only=False, pages=1, popularity=1.0, size=250):
    pkg_list = []
    for page in range(0, pages):
        from_value = page * 250
        registry_search_url = f"{config.NPM_SERVER}/-/v1/search?popularity={popularity}&size={size}&from={from_value}"
        if insecure_only:
            registry_search_url = f"{registry_search_url}&text=is:insecure"
        elif unstable_only:
            registry_search_url = f"{registry_search_url}&text=is:unstable"
        elif keywords:
            registry_search_url = f"{registry_search_url}&text=keywords:{','.join(keywords)}"
        else:
            registry_search_url = f"{registry_search_url}&text=not:insecure"
        try:
            r = httpclient.get(
                url=registry_search_url,
                follow_redirects=True,
                timeout=config.request_timeout_sec,
            )
            result = r.json()
            if result and not r.is_error and result.get("objects"):
                for aobj in result.get("objects"):
                    if aobj and aobj.get("package"):
                        package = aobj.get("package")
                        flags = aobj.get("flags", {})
                        name = package.get("name")
                        if name.startswith("@types/"):
                            continue
                        is_pkg_insecure = True if flags.get("insecure", 0) == 1 else False
                        pkg_list.append(
                            {
                                "name": name,
                                "version": package.get("version"),
                                "purl": f'pkg:npm/{package.get("name").replace("@", "%40")}@{package.get("version")}',
                                "insecure": is_pkg_insecure,
                                "unstable": flags.get("unstable", False)
                            }
                        )
        except Exception as e:
            print(e)
            pass
    return pkg_list


def get_npm_download_stats(name, period="last-year"):
    """
    Method to download npm stats

    :param name: Package name
    :param period: Stats period
    """
    stats_url = f"https://api.npmjs.org/downloads/point/{period}/{name}"
    try:
        r = httpclient.get(
            url=stats_url,
            follow_redirects=True,
            timeout=config.request_timeout_sec,
        )
        return r.json()
    except Exception:
        return {}


def npm_pkg_risk(pkg_metadata, is_private_pkg, scope, pkg):
    """
    Calculate various npm package risks based on the metadata from npm. The
    keys in the risk_metrics dict is based on the parameters specified in
    config.py and has a _risk suffix. Eg: config.pkg_min_versions would
    result in a boolean pkg_min_versions_risk and pkg_min_versions_value

    :param pkg_metadata: A dict containing the metadata of the npm package.
    :param is_private_pkg: Boolean to indicate if this package is private
    :param scope: Package scope
    :param pkg: Package object

    :return: A dict containing the calculated risks and score.
    """
    # Some default values to ensure the structure is non-empty
    risk_metrics = {
        "pkg_deprecated_risk": False,
        "pkg_version_deprecated_risk": False,
        "pkg_version_missing_risk": False,
        "pkg_includes_binary_risk": False,
        "pkg_min_versions_risk": False,
        "created_now_quarantine_seconds_risk": False,
        "latest_now_max_seconds_risk": False,
        "mod_create_min_seconds_risk": False,
        "pkg_min_maintainers_risk": False,
        "pkg_node_version_risk": False,
        "pkg_private_on_public_registry_risk": False,
    }
    # Is the private package available publicly? Dependency confusion.
    if is_private_pkg and pkg_metadata:
        risk_metrics["pkg_private_on_public_registry_risk"] = True
        risk_metrics["pkg_private_on_public_registry_value"] = 1
    versions = pkg_metadata.get("versions", {})
    latest_version = pkg_metadata.get("dist-tags", {}).get("latest")
    engines_block_dict = versions.get(latest_version, {}).get("engines", {})
    # Check for scripts block
    scripts_block_dict = versions.get(latest_version, {}).get("scripts", {})
    bin_block_dict = versions.get(latest_version, {}).get("bin", {})
    theversion = None
    if pkg:
        if pkg.get("version"):
            theversion = versions.get(pkg.get("version"), {})
            # Check if the version exists in the registry
            if not theversion:
                risk_metrics["pkg_version_missing_risk"] = True
                risk_metrics["pkg_version_missing_value"] = 1
        # Proceed with the rest of checks using the latest version
        if not theversion:
            theversion = versions.get(latest_version, {})
        # Get the version specific engines and scripts block
        if theversion.get("engines"):
            engines_block_dict = theversion.get("engines")
        if theversion.get("scripts"):
            scripts_block_dict = theversion.get("scripts")
        if theversion.get("bin"):
            bin_block_dict = theversion.get("bin")
        # Check if there is any binary downloaded and offered
        if theversion.get("binary"):
            risk_metrics["pkg_includes_binary_risk"] = True
            risk_metrics["pkg_includes_binary_value"] = 1
            # Capture the remote host
            if theversion["binary"].get("host"):
                risk_metrics["pkg_includes_binary_info"] = (
                    f'Host: {theversion["binary"].get("host")}\nBinary: {theversion["binary"].get("module_name")}'
                )
            # For some packages,
            elif theversion["binary"].get("napi_versions"):
                if theversion.get("repository", {}).get("url"):
                    risk_metrics["pkg_includes_binary_info"] = (
                        f'Repository: {theversion.get("repository").get("url")}'
                    )
                elif theversion.get("homepage"):
                    risk_metrics["pkg_includes_binary_info"] = (
                        f'Homepage: {theversion.get("homepage")}'
                    )
        elif bin_block_dict and maybe_binary_npm_package(pkg.get("name")):
            # See #317
            risk_metrics["pkg_includes_binary_risk"] = True
            risk_metrics["pkg_includes_binary_value"] = len(
                bin_block_dict.keys()
            )
            bin_block_desc = ""
            for k, v in bin_block_dict.items():
                bin_block_desc = f"{bin_block_desc}\n{k}: {v}"
            if bin_block_desc:
                risk_metrics["pkg_includes_binary_info"] = (
                    f"Binary commands:{bin_block_desc}"
                )
        # Look for slsa attestations
        if theversion.get("dist", {}).get("attestations") and theversion.get(
            "dist", {}
        ).get("signatures"):
            attestations = theversion.get("dist").get("attestations")
            signatures = theversion.get("dist").get("signatures")
            if (
                attestations.get("url").startswith(
                    "https://registry.npmjs.org/"
                )
                and attestations.get("provenance", {}).get("predicateType", "")
                == "https://slsa.dev/provenance/v1"
            ):
                risk_metrics["pkg_attested_check"] = True
                risk_metrics["pkg_attested_value"] = len(signatures)
                risk_metrics["pkg_attested_info"] = "\n".join(
                    [sig.get("keyid") for sig in signatures]
                )
        # In some packages like biomejs, there would be no binary section
        # case 1: optional dependencies section might have a bunch of packages for each os
        # case 2: prebuild, prebuild-install, prebuildify in dependencies
        # case 3: there could be a libc attribute
        # case 4: fileCount <= 2 and size > 20 MB
        if not theversion.get("binary"):
            binary_count = 1
            if theversion.get("bin") and isinstance(theversion.get("bin"), dict):
                binary_count = max(len(theversion.get("bin", {}).keys()), 1)
            for opkg in theversion.get("optionalDependencies", {}).keys():
                if (
                    "linux" in opkg
                    or "darwin" in opkg
                    or "win32" in opkg
                    or "arm64" in opkg
                    or "musl" in opkg
                ):
                    risk_metrics["pkg_includes_binary_risk"] = True
                    risk_metrics["pkg_includes_binary_value"] = binary_count
                    break
            # Eg: pkg:npm/zeromq@6.0.0-beta.19
            dev_deps = list(theversion.get("devDependencies", {}).keys())
            direct_deps = list(theversion.get("dependencies", {}).keys())
            if "prebuild" in " ".join(dev_deps) or "prebuild" in " ".join(
                direct_deps
            ):
                risk_metrics["pkg_includes_binary_risk"] = True
                risk_metrics["pkg_includes_binary_value"] = binary_count
            if not risk_metrics.get("pkg_includes_binary_risk"):
                if theversion.get("libc"):
                    risk_metrics["pkg_includes_binary_risk"] = True
                    risk_metrics["pkg_includes_binary_value"] = len(
                        theversion.get("libc", [])
                    )
                elif (
                    theversion.get("dist", {}).get("fileCount", 0) <= 2
                    and theversion.get("dist", {}).get("unpackedSize")
                    and (
                        theversion.get("dist").get("unpackedSize", 0)
                        / (1000 * 1000)
                    )
                    > 20
                ):
                    risk_metrics["pkg_includes_binary_risk"] = True
                    risk_metrics["pkg_includes_binary_value"] = 1
    is_deprecated = (
        versions.get(latest_version, {}).get("deprecated", None) is not None
    )
    is_version_deprecated = (
        True if theversion and theversion.get("deprecated") else False
    )
    # Is the package deprecated
    if is_deprecated:
        risk_metrics["pkg_deprecated_risk"] = True
        risk_metrics["pkg_deprecated_value"] = 1
    elif is_version_deprecated:
        risk_metrics["pkg_version_deprecated_risk"] = True
        risk_metrics["pkg_version_deprecated_value"] = 1
        # The deprecation reason for a specific version are often useful
        risk_metrics["pkg_version_deprecated_info"] = theversion.get(
            "deprecated"
        )
    scripts_block_list = []
    # There are some packages on npm with incorrectly configured scripts
    # block Good news is that the install portion would only for if the
    # scripts block is an object/dict
    if isinstance(scripts_block_dict, dict):
        scripts_block_list = [
            block
            for block in scripts_block_dict.keys()
            if block in ("preinstall", "postinstall", "prebuild")
        ]
        # Detect the use of prebuild-install
        # https://github.com/prebuild/prebuild-install
        # https://github.com/prebuild/prebuildify
        if not risk_metrics.get("pkg_includes_binary_risk"):
            if scripts_block_dict.get("prebuild", "").startswith("prebuild"):
                risk_metrics["pkg_includes_binary_risk"] = True
                risk_metrics["pkg_includes_binary_value"] = 1
    # If the package has fewer than minimum number of versions
    if len(versions) < config.pkg_min_versions:
        risk_metrics["pkg_min_versions_risk"] = True
        risk_metrics["pkg_min_versions_value"] = len(versions)
    # Time related checks
    time_info = pkg_metadata.get("time", {})
    modified = time_info.get("modified", "").replace("Z", "")
    created = time_info.get("created", "").replace("Z", "")
    if not modified and pkg_metadata.get("mtime"):
        modified = pkg_metadata.get("mtime").replace("Z", "")
    if not created and pkg_metadata.get("ctime"):
        created = pkg_metadata.get("ctime").replace("Z", "")
    latest_version_time = time_info.get(latest_version, "").replace("Z", "")
    if time_info and modified and created and latest_version_time:
        modified_dt = datetime.fromisoformat(modified)
        created_dt = datetime.fromisoformat(created)
        latest_version_time_dt = datetime.fromisoformat(latest_version_time)
        mod_create_diff = modified_dt - created_dt
        latest_now_diff = datetime.now() - latest_version_time_dt
        created_now_diff = datetime.now() - created_dt
        risk_metrics = compute_time_risks(
            risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
        )

    # Maintainers count related risk. Ignore packages that are past
    # quarantine period
    maintainers = pkg_metadata.get("maintainers", [])
    if len(maintainers) < config.pkg_min_maintainers and risk_metrics.get(
        "created_now_quarantine_seconds_risk"
    ):
        risk_metrics["pkg_min_maintainers_risk"] = True
        risk_metrics["pkg_min_maintainers_value"] = len(maintainers)
        # Check for install scripts risk only for those packages with
        # maintainers risk
        if scripts_block_list:
            risk_metrics["pkg_install_scripts_risk"] = True
            risk_metrics["pkg_install_scripts_value"] = len(scripts_block_list)

    # Users count related risk. Ignore packages that are past quarantine period
    users = pkg_metadata.get("users", [])
    if (
        users
        and len(users) < config.pkg_min_users
        and risk_metrics.get("created_now_quarantine_seconds_risk")
    ):
        risk_metrics["pkg_min_users_risk"] = True
        risk_metrics["pkg_min_users_value"] = len(users)
    # Node engine version There are packages with incorrect node engine
    # specification which we can ignore for now
    if (
        engines_block_dict
        and isinstance(engines_block_dict, dict)
        and engines_block_dict.get("node")
        and isinstance(engines_block_dict.get("node"), str)
    ):
        node_version_spec = engines_block_dict.get("node")
        node_version = (
            node_version_spec.replace(">= ", "")
            .replace(">=", "")
            .replace("> ", "")
            .replace(">", "")
            .replace("~ ", "")
            .replace("~", "")
            .split(" ")[0]
        )
        for ver in config.pkg_node_version.split(","):
            if node_version.startswith(ver):
                risk_metrics["pkg_node_version_risk"] = True
                risk_metrics["pkg_node_version_value"] = 1
                break
    # Add package scope related weight
    if scope:
        risk_metrics[f"pkg_{scope}_scope_risk"] = True
        risk_metrics[f"pkg_{scope}_scope_value"] = 1

    risk_metrics["risk_score"] = calculate_risk_score(risk_metrics)
    return risk_metrics


def maybe_binary_npm_package(name: str) -> bool:
    """
    Check if a package might be a binary by checking the naming conventions.

    :param name: Packagename
    :returns: boolean
    """
    if not name:
        return False
    for bin_suffix in config.NPM_BINARY_PACKAGES_SUFFIXES:
        if name.endswith(bin_suffix):
            return True
    return False



import os
import sys
from os.path import dirname, exists, join


def resource_path(relative_path):
    """

    :param relative_path:
    :return:
    """
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = dirname(__file__)
    return join(base_path, relative_path)


license_data_dir = resource_path(
    join(
        "..",
        "..",
        "vendor",
        "choosealicense.com",
        "_licenses",
    )
)
spdx_license_list = resource_path(
    join(
        "..",
        "..",
        "vendor",
        "spdx",
        "json",
        "licenses.json",
    )
)
if not exists(license_data_dir):
    license_data_dir = resource_path(
        join(
            "vendor",
            "choosealicense.com",
            "_licenses",
        )
    )
    spdx_license_list = resource_path(
        join(
            "vendor",
            "spdx",
            "json",
            "licenses.json",
        )
    )


# Default ignore list
ignore_directories = [
    ".git",
    ".svn",
    ".mvn",
    ".idea",
    "dist",
    "bin",
    "obj",
    "backup",
    "docs",
    "tests",
    "test",
    "tmp",
    "report",
    "reports",
    "node_modules",
    ".terraform",
    ".serverless",
    "venv",
    "examples",
    "tutorials",
    "samples",
    "migrations",
    "db_migrations",
    "unittests",
    "unittests_legacy",
    "stubs",
    "mock",
    "mocks",
]


def get_float_from_env(name, default):
    """
    Retrieves a value from an environment variable and converts it to a
    float. If the value cannot be converted to a float, it returns the
    default value provided.

    :param name:
    :param default:
    :return:
    """
    value = os.getenv(name.upper(), default)
    try:
        value = float(value)
    except ValueError:
        value = default
    return value


def get_int_from_env(name, default):
    """
    Retrieves a value from an environment variable and converts it to an
    integer. If the value cannot be converted to an integer, it returns the
    default value provided.

    :param name:
    :param default:
    """
    return int(get_float_from_env(name, default))


NPM_SERVER = "https://registry.npmjs.org"
npm_app_info = {"name": "owasp-depscan", "version": "6.0.0"}

PYPI_SERVER = "https://pypi.org/pypi"

CARGO_SERVER = "https://crates.io/api/v1/crates"

# Use the env variable VDB_DATABASE_URL=ghcr.io/appthreat/vdbxz-app:v6.5.x for app-only database
vdb_database_url = os.getenv("VDB_DATABASE_URL", "ghcr.io/appthreat/vdbxz:v6.5.x")

# Larger 10 year database
vdb_10y_database_url = os.getenv(
    "VDB_10Y_DATABASE_URL", "ghcr.io/appthreat/vdbxz-10y:v6.5.x"
)

if os.getenv("USE_VDB_10Y", "") in ("true", "1"):
    vdb_database_url = vdb_10y_database_url

# How old vdb can be before it gets re-downloaded. 48 hours.
VDB_AGE_HOURS = get_int_from_env("VDB_AGE_HOURS", 48)

# Package risk scoring using a simple weighted formula with no backing
# research All parameters and their max value and weight can be overridden
# using environment variables

# Some constants and defaults
SECONDS_IN_DAY = 24 * 60 * 60
SECONDS_IN_HOUR = 60 * 60
DEFAULT_MAX_VALUE = 100
DEFAULT_WEIGHT = 1

# Package should have at least 3 versions
pkg_min_versions = get_float_from_env("pkg_min_versions", 3)
pkg_min_versions_max = get_float_from_env("pkg_min_versions_max", 100)
pkg_min_versions_weight = get_float_from_env("pkg_min_versions_weight", 2)

# At least 12 hours difference between the creation and modified time
mod_create_min_seconds = get_float_from_env(
    "mod_create_min_seconds", 12 * SECONDS_IN_HOUR
)
mod_create_min_seconds_max = get_float_from_env(
    "mod_create_min_seconds_max", 1000 * SECONDS_IN_DAY
)
mod_create_min_seconds_weight = get_float_from_env("mod_create_min_seconds_weight", 1)

# At least 12 hours difference between the latest version and the current time
latest_now_min_seconds = get_float_from_env(
    "latest_now_min_seconds", 12 * SECONDS_IN_HOUR
)
latest_now_min_seconds_max = get_float_from_env(
    "latest_now_min_seconds_max", 1000 * SECONDS_IN_DAY
)
latest_now_min_seconds_weight = get_float_from_env("latest_now_min_seconds_weight", 0.5)

# Time period after which certain risks can be considered safe. Quarantine
# period For eg: Packages that are over 1 year old
created_now_quarantine_seconds = get_float_from_env(
    "created_now_quarantine_seconds", 365 * SECONDS_IN_DAY
)
created_now_quarantine_seconds_max = get_float_from_env(
    "created_now_quarantine_seconds_max", 365 * SECONDS_IN_DAY
)
created_now_quarantine_seconds_weight = get_float_from_env(
    "created_now_quarantine_seconds_weight", 0.5
)

# Max package age - 6 years
latest_now_max_seconds = get_float_from_env(
    "latest_now_max_seconds", 6 * 365 * SECONDS_IN_DAY
)
latest_now_max_seconds_max = get_float_from_env(
    "latest_now_max_seconds_max", 6 * 365 * SECONDS_IN_DAY
)
latest_now_max_seconds_weight = get_float_from_env("latest_now_max_seconds_weight", 0.5)

# Package should have at least 2 maintainers
pkg_min_maintainers = get_float_from_env("pkg_min_maintainers", 2)
pkg_min_maintainers_max = get_float_from_env("pkg_min_maintainers_max", 20)
pkg_min_maintainers_weight = get_float_from_env("pkg_min_maintainers_weight", 2)

# Package should have at least 2 users
pkg_min_users = get_float_from_env("pkg_min_users", 2)
pkg_min_users_max = get_float_from_env("pkg_min_users_max", 20)
pkg_min_users_weight = get_float_from_env("pkg_min_users_weight", 0.25)

# Package with install scripts (npm)
pkg_install_scripts_max = get_float_from_env("pkg_install_scripts_max", 0)
pkg_install_scripts_weight = get_float_from_env("pkg_install_scripts_weight", 2)

# Node version risk
pkg_node_version = os.getenv("pkg_node_version".upper(), "0.,4,6,8,10,12")
pkg_node_version_max = get_float_from_env("pkg_node_version_max", 16)
pkg_node_version_weight = get_float_from_env("pkg_node_version_weight", 0.5)

# Package deprecated
pkg_deprecated_weight = get_float_from_env("pkg_deprecated_weight", 2)
pkg_deprecated_max = get_float_from_env("pkg_deprecated_max", 0)

# Package version deprecated
pkg_version_deprecated_weight = get_float_from_env("pkg_version_deprecated_weight", 2)
pkg_version_deprecated_max = get_float_from_env("pkg_version_deprecated_max", 0)

# Package version missing
pkg_version_missing_weight = get_float_from_env("pkg_version_missing_weight", 2)
pkg_version_missing_max = get_float_from_env("pkg_version_missing_max", 0)

# Package includes binary
pkg_includes_binary_weight = get_float_from_env("pkg_includes_binary_weight", 2)
pkg_includes_binary_max = get_float_from_env("pkg_includes_binary_max", 0)

# Package has attestation
pkg_attested_weight = get_float_from_env("pkg_attested_weight", -2)
pkg_attested_max = get_float_from_env("pkg_attested_max", 0)

# Package dependency confusion
pkg_private_on_public_registry_weight = get_float_from_env(
    "pkg_private_on_public_registry_weight", 4
)
pkg_private_on_public_registry_max = get_float_from_env(
    "pkg_private_on_public_registry_max", 1
)

# Package scope related weight
pkg_required_scope_weight = get_float_from_env("pkg_required_scope_weight", 4.0)
pkg_optional_scope_weight = get_float_from_env("pkg_optional_scope_weight", 0.5)
pkg_excluded_scope_weight = get_float_from_env("pkg_excluded_scope_weight", 0)
pkg_required_scope_max = get_float_from_env("pkg_required_scope_max", 1)
pkg_optional_scope_max = get_float_from_env("pkg_optional_scope_max", 1)
pkg_excluded_scope_max = get_float_from_env("pkg_excluded_scope_max", 1)

total_weight = (
    pkg_min_versions_weight
    + mod_create_min_seconds_weight
    + latest_now_min_seconds_weight
    + latest_now_max_seconds_weight
    + created_now_quarantine_seconds_weight
    + pkg_min_maintainers_weight
    + pkg_min_users_weight
    + pkg_install_scripts_weight
    + pkg_node_version_weight
    + pkg_required_scope_weight
    + pkg_optional_scope_weight
    + pkg_deprecated_weight
    + pkg_version_deprecated_weight
    + pkg_version_missing_weight
    + pkg_includes_binary_weight
    + pkg_private_on_public_registry_weight
)


# Package max risk score. All packages above this level will be reported
pkg_max_risk_score = get_float_from_env("pkg_max_risk_score", 0.5)

# Default request timeout
request_timeout_sec = get_int_from_env("request_timeout_sec", 20)

# Number of api failures that would stop the risk audit completely
max_request_failures = get_int_from_env("max_request_failures", 5)

# Universal scan
UNIVERSAL_SCAN_TYPE = "universal"

max_reachable_explanations = get_int_from_env("max_reachable_explanations", 20)

# How many explanations for a given combination of purls
max_purls_reachable_explanations = get_int_from_env(
    "max_purls_reachable_explanations", 3
)
max_source_reachable_explanations = get_int_from_env(
    "max_source_reachable_explanations", 2
)
max_sink_reachable_explanations = get_int_from_env("max_sink_reachable_explanations", 2)

max_purl_per_flow = get_int_from_env("max_purl_per_flow", 8)
max_flows_per_prompt = get_int_from_env("max_flows_per_prompt", 8)

RUBY_PLATFORM_MARKERS = [
    "-x86_64",
    "-x86",
    "-x64",
    "-aarch",
    "-arm",
    "-ruby",
    "-universal",
    "-java",
    "-truffle",
]

# List of suffixes used by npm packages to indicate binary versions.
# This could be replaced with a better heuristics or lookup database in the future.
NPM_BINARY_PACKAGES_SUFFIXES = ("-prebuilt",)

DEPSCAN_DEFAULT_VDR_FILE = os.getenv(
    "DEPSCAN_DEFAULT_VDR_FILE", "depscan-universal.vdr.json"
)

COMMON_CHECK_TAGS = (
    "validation",
    "encode",
    "encrypt",
    "sanitize",
    "authentication",
    "authorization",
)



from datetime import datetime, timezone

from depscan.lib import config
from depscan.lib.package_query.pkg_query import compute_time_risks, calculate_risk_score
from semver import Version

def set_binary_risks(risk_metrics, current_version, latest_version):
    """
    If current version has bin_names. then we should set "pkg_includes_binary_risk" as True.
    and add the number of bin_names to the "pkg_includes_binary_value" key.
    """
    version = current_version if current_version else latest_version
    bin_names = version.get('bin_names', [])
    risk_metrics["pkg_includes_binary_risk"] = True if len(bin_names)>0 else False
    risk_metrics["pkg_includes_binary_value"] = len(bin_names)


def cargo_pkg_risk(pkg_metadata, is_private_pkg, scope, pkg):
    """
    Calculate various package risks based on the metadata from cargo.
    """
    risk_metrics = {
        "pkg_deprecated_risk": False,
        "pkg_version_deprecated_risk": False,
        "pkg_version_missing_risk": False,
        "pkg_includes_binary_risk": False,
        "pkg_min_versions_risk": False,
        "created_now_quarantine_seconds_risk": False,
        "latest_now_max_seconds_risk": False,
        "mod_create_min_seconds_risk": False,
        "pkg_min_maintainers_risk": False,
        "pkg_node_version_risk": False,
        "pkg_private_on_public_registry_risk": False,
    }
    versions_list = pkg_metadata.get("versions", [])
    versions_dict = {
        crate_version.get('num'): crate_version
        for crate_version in versions_list}
    versions_nums = [
        crate_version.get('num')
        for crate_version in versions_list]

    is_deprecated = versions_list[0].get("yanked")
    is_version_deprecated = False
    info = pkg_metadata.get("crate", {})
    if not is_deprecated and pkg and pkg.get("version"):
        theversion = versions_dict.get(pkg.get("version"), {})
        if isinstance(theversion, dict) and len(theversion) > 0:
            theversion = theversion.get('num')
        elif theversion and theversion.get("yanked"):
            is_version_deprecated = True
        # Check if the version exists in the registry
        if not theversion:
            risk_metrics["pkg_version_missing_risk"] = True
            risk_metrics["pkg_version_missing_value"] = 1

    pkg_description = info.get("description", "").lower()
    if not is_deprecated and (
        "is deprecated" in pkg_description
        or "no longer maintained" in pkg_description
    ):
        is_deprecated = True
    latest_deprecated = False

    try:
        first_version_num = min(
            versions_nums,
            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
        )
        latest_version_num = max(
            versions_nums,
            key=lambda x: Version.parse(x, optional_minor_and_patch=True),
        )
    except (ValueError, TypeError):
        first_version_num = versions_nums[-1]
        latest_version_num = versions_nums[0]
    
    first_version = versions_dict[first_version_num]
    latest_version = versions_list[latest_version_num]

    # Is the private package available publicly? Dependency confusion.
    if is_private_pkg and pkg_metadata:
        risk_metrics["pkg_private_on_public_registry_risk"] = True
        risk_metrics["pkg_private_on_public_registry_value"] = 1

    # If the package has fewer than minimum number of versions
    if len(versions_list):
        if len(versions_list) < config.pkg_min_versions:
            risk_metrics["pkg_min_versions_risk"] = True
            risk_metrics["pkg_min_versions_value"] = len(versions_list)
        # Check if the latest version is deprecated
        if latest_version and latest_version.get("yanked"):
            latest_deprecated = True

    # Created and modified time related checks
    if first_version and latest_version:
        created = first_version.get("created_at")
        modified = latest_version.get("updated_at")
        if created and modified:
            modified_dt = datetime.fromisoformat(modified)
            created_dt = datetime.fromisoformat(created)
            mod_create_diff = modified_dt - created_dt
            latest_now_diff = datetime.now(timezone.utc) - modified_dt
            created_now_diff = datetime.now(timezone.utc) - created_dt
            risk_metrics = compute_time_risks(
                risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
            )

    # Is the package deprecated
    if is_deprecated or latest_deprecated:
        risk_metrics["pkg_deprecated_risk"] = True
        risk_metrics["pkg_deprecated_value"] = 1
    elif is_version_deprecated:
        risk_metrics["pkg_version_deprecated_risk"] = True
        risk_metrics["pkg_version_deprecated_value"] = 1
    # Add package scope related weight
    if scope:
        risk_metrics[f"pkg_{scope}_scope_risk"] = True
        risk_metrics[f"pkg_{scope}_scope_value"] = 1

    set_binary_risks(risk_metrics, versions_dict.get(pkg.get("version"), {}), latest_version)

    risk_metrics["risk_score"] = calculate_risk_score(risk_metrics)
    return risk_metrics



import logging
import os
import random
import re
import sys

from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.logging import RichHandler
from rich.theme import Theme


class CustomHighlighter(RegexHighlighter):
    base_style = "depscan."
    highlights = [
        r"(?P<method>([\w-]+\.)+[\w-]+[^<>:(),]?)",
        r"(?P<path>(\w+\/.*\.[\w:]+))",
        r"(?P<params>[(]([\w,-]+\.)+?[\w-]+[)]$)",
        r"(?P<opers>(unresolvedNamespace|unresolvedSignature|init|operators|operator|clinit))",
    ]


custom_theme = Theme(
    {
        "depscan.path": "#7c8082",
        "depscan.params": "#5a7c90",
        "depscan.opers": "#7c8082",
        "depscan.method": "#FF753D",
        "info": "#5A7C90",
        "warning": "#FF753D",
        "danger": "bold red",
    }
)

IS_CI = os.getenv("CI") or os.getenv("CONTINUOUS_INTEGRATION")

console = Console(
    log_time=False,
    log_path=False,
    theme=custom_theme,
    color_system=os.getenv("CONSOLE_COLOR_SCHEME", "256"),
    width=140 if IS_CI else None,
    highlight=not IS_CI,
    tab_size=2,
    highlighter=CustomHighlighter(),
    record=sys.platform == "win32",
    emoji=os.getenv("DISABLE_CONSOLE_EMOJI", "") not in ("true", "1"),
)

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(
            console=console,
            markup=True,
            show_path=False,
            enable_link_path=False,
        )
    ],
)
LOG = logging.getLogger(__name__)

# Set logging level
if os.getenv("SCAN_DEBUG_MODE") == "debug":
    LOG.setLevel(logging.DEBUG)

DEBUG = logging.DEBUG
for log_name, log_obj in logging.Logger.manager.loggerDict.items():
    if not log_name.startswith("depscan"):
        log_obj.disabled = True

SPINNER = os.getenv(
    "DEPSCAN_SPINNER",
    random.choice(
        [
            "pong",
            "arrow3",
            "bouncingBall",
            "dots2",
            "material",
            "shark",
            "simpleDotsScrolling",
            "toggle9",
        ]
    ),
)

# Support for thought logging
tlogger = None


def thought_log(s):
    if s and tlogger and tlogger.isEnabledFor(DEBUG):
        s = re.sub(r"([.!?])?$", ".", s)
        tlogger.debug(s)


def thought_begin():
    thought_log("<think>")


def thought_end():
    thought_log("</think>")


if os.getenv("DEPSCAN_THINK_MODE", "") in ("true", "1"):
    tlogger = logging.getLogger("depscan_thoughts")
    tlogger.setLevel(DEBUG)
    file_handler = logging.FileHandler(
        os.getenv(
            "DEPSCAN_THOUGHT_LOG", os.path.join(os.getcwd(), "depscan-thoughts.log")
        )
    )
    file_handler.setLevel(DEBUG)
    formatter = logging.Formatter("%(message)s")
    file_handler.setFormatter(formatter)
    tlogger.addHandler(file_handler)
